;; -*- Mode: Scheme -*-
;;   Machine description for GNU compiler,
;;   for Texas Instruments msp430 MCUs
;;   Copyright (C) 2001, 2002, 2003 Free Software Foundation, Inc.
;;   Contributed by Dmitry Diky <diwil@mail.ru>
;;	 GCC 4.x port by Ivan Shcherbakov <mspgcc@sysprogs.org>

;; This work is partially financed by the European Commission under the
;; Framework 6 Information Society Technologies Project
;; "Wirelessly Accessible Sensor Populations (WASP)".

;

; This file is part of GCC.

;; GCC is free software; you can redistribute it and/or modify
;; it under the terms of the GNU General Public License as published by
;; the Free Software Foundation; either version 2, or (at your option)
;; any later version.

;; GCC is distributed in the hope that it will be useful,
;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;; GNU General Public License for more details.

;; You should have received a copy of the GNU General Public License
;; along with GCC; see the file COPYING.  If not, write to
;; the Free Software Foundation, 59 Temple Place - Suite 330,
;; Boston, MA 02111-1307, USA.

(include "predicates.md")

;; Special characters after '%':
;;  A  No effect (add 0).
;;  B  Add 1 to REG number, 2 to MEM address or CONST_INT.
;;  C	   2		    4
;;  D	   3		    6
;;  E  adds nothing to reg but used only with (mem:hi (reg:hi))
;;  F  no trim array
;;  M  Add 0 to address if using stack pointer
;;  N  Add 2 to address if using stack pointer
;;  Extra constarains:
;;  P  hardware constants: -1,0,+1,+2,+4,+8
;;  Q  Indexed destination register as X(Rn)
;;  R  Indexed source register as @Rn+
;;  S  Symbol reference for 'C' like: a = *b;
;;

;; Unspec usage:
;; 3 - strlen
;; 0  - addc_reg
;; 5  - addc_any
;; 1  - bittest_lo
;; 2  - bittest_hi
;; 6  - bittest
;; 4  - swpb
;; 7  - bittest_b
;; 8  - move SF to SI with no conversion

(define_constants
  [(REG_PC			0)
   (REG_SP			1)
   (UNSPEC_EINT		100)
   (UNSPEC_DINT		101)
   (UNSPEC_EXPLICIT_BR 102)
   (UNSPEC_PUSH_SREG 103)
   (UNSPEC_SAVE_PC_TO_REG 104)
   (UNSPEC_LOAD_SP	105)
   (UNSPEC_PROLOGUE_PUSH 106)
   (UNSPEC_POP_R2	107)
   ]
)

;; Condition code settings.


(define_attr "cc" "none,set_czn,set_zn,set_n,compare,clobber,further,oper,cbranch"
  (const_string "none"))

(define_attr "type" "branch,branch1,arith"
  (const_string "arith"))

(define_attr "msp430_has_hwmul" "yes,no"
  (const (if_then_else (symbol_ref "MSP430_HAS_HWMUL_INTERNAL")
		       (const_string "yes")
		       (const_string "no"))))

(define_attr "msp430_noint_hwmul" "" (symbol_ref "MSP430_NOINT_HWMUL"))

;; The size of instructions in bytes.
;; XXX may depend from "cc"

;; for confitional branches
(define_attr "length" ""
  (cond [(eq_attr "type" "branch")
         (if_then_else  (and (ge (minus (pc) (match_dup 0)) 
	 				(const_int -508))
                             (le (minus (pc) (match_dup 0)) 
			     		(const_int  508)))
                        (const_int 1)
			(const_int 2))]
	(const_int 2)
))


;;========================================================================
;;  PUSH/POP helper functions
;;


(define_insn "*pushqi_pre_mod"
[(set (mem:QI (pre_modify:HI (reg:HI 1)
                (plus:HI (reg:HI 1) (const_int -2))))
	(match_operand:QI 0 "general_operand" "rim"))]
""
"* return msp430_pushqi(insn, operands, NULL);"
[(set_attr "length" "2")])

(define_insn "*pushqi"
  [(set (match_operand:QI 1 "push_operand" "=<")
;;  [(set (mem:QI (post_dec (reg:HI 1)))	;PRE_DEC!
        (match_operand:QI 0 "general_operand" "rim"))]
  ""
  "* return msp430_pushqi(insn, operands, NULL);"
  [(set_attr "length" "2")])

(define_insn "*pushhi"
  [(set (match_operand:HI 1 "push_operand" "=<")
        (match_operand:HI 0 "general_no_elim_operand" "rim"))]
  ""
  "* return msp430_pushhi(insn, operands, NULL);"
  [(set_attr "length" "2")])

(define_insn "pushhi_prologue"
  [(set (match_operand:HI 1 "push_operand" "=<")
        (match_operand:HI 0 "general_operand" "rim"))
	(unspec_volatile:HI [(const_int 0)] UNSPEC_PROLOGUE_PUSH)]
  ""
  "* return msp430_pushhi(insn, operands, NULL);"
  [(set_attr "length" "2")])

  
;(define_insn_and_split "*pushhi"
;  [(set (match_operand:HI 1 "push_operand" "=<")
;        (match_operand:HI 0 "general_operand" "rim"))]
;  ""
;  "* return msp430_pushhi(insn, operands, NULL);"
;  "(GET_CODE(operands[0]) == REG) && ((REGNO(operands[0]) == REG_SP) || (operands[0] == arg_pointer_rtx) || (operands[0] == frame_pointer_rtx) || IN_RANGE (REGNO (operands[0]), FIRST_PSEUDO_REGISTER, LAST_VIRTUAL_REGISTER))"
;  [
;	(set (match_dup 2) (match_dup 0))
;	(set (match_dup 1) (match_dup 2))
;  ]
;  "operands[2] = gen_rtx_reg(SImode);"
;  [(set_attr "length" "2")]
; )

 

(define_insn "*pushsi"
  [(set (match_operand:SI 1 "push_operand" "=<")
;;  [(set (mem:SI (post_dec (reg:HI 1))) ;PRE_DEC!
        (match_operand:SI 0 "general_operand" "rmi"))]
  ""
  "* return msp430_pushsisf(insn, operands, NULL);"
  [(set_attr "length" "4")])


(define_insn "*pushdi"
  [(set (match_operand:DI 1 "push_operand" "=<")
;;  [(set (mem:DI (post_dec (reg:HI 1)))	;PRE_DEC!
        (match_operand:DI 0 "general_operand" "rmi"))]
  ""
  "* return msp430_pushdi(insn, operands, NULL);"
  [(set_attr "length" "8")])


(define_insn "*pushsf"
  [(set (match_operand:SF 1 "push_operand" "=<")
;;  [(set (mem:SF (post_dec (reg:HI 1)))	;PRE+DEC
        (match_operand:SF 0 "general_operand" "rmi"))]
  ""
  "* return msp430_pushsisf(insn, operands, NULL);"
  [(set_attr "length" "4")])
  
(define_insn "push_sreg"
  [(unspec_volatile:HI [(const_int 0)] UNSPEC_PUSH_SREG)
   (set (reg:HI REG_SP) (minus:HI 
                           (reg:HI REG_SP)
                           (const_int 2)))]
  ""
  "push r2"
  [(set_attr "length" "2")])


;; =========================================== POP register instruction ====================================

;; The POP instruction is used only by expand_epilogue() function.
;; As present implementation of DWARF2 CFA generator (gcc 4.4.2) does
;; not recognize pop_operand-based or post_inc-based operations,
;; the short and neat definition (first one) cannot be used.
;; Instead, we use a longer PARALLEL format to define the action of
;; "pop", so the DWARF2 generator will be happy.

;;(define_insn "pophi_reg"
;;  [(set (match_operand:HI 0 "register_operand" "=r")
;;		(match_operand:HI 1 "pop_operand" ">"))]
;;  ""
;;  "pop %0"
;;  [(set_attr "length" "1")])

(define_insn "pophi_reg"
  [(parallel [
    (set (reg:HI REG_SP) (plus:HI (reg:HI REG_SP) (const_int 2)))				;;This goes first, so the DWARF2 generator can recognize it
	(set (match_operand:HI 0 "register_operand" "=r") (mem:HI (reg:HI REG_SP)))	;;This goes second with RTX_FRAME_RELATED_P() set to 0
	]
	)]
  ""
  "pop\t%0"
  [(set_attr "length" "1")])

(define_insn "pop_r2"
  [(set (reg:HI REG_SP) (plus:HI (reg:HI REG_SP) (const_int 2)))
	(unspec_volatile:HI [(const_int 0)] UNSPEC_POP_R2)]
  ""
  "pop\tr2"
  [(set_attr "length" "2")])

(define_insn "save_pc_to_reg"
  [(unspec_volatile:HI [(const_int 0)] UNSPEC_SAVE_PC_TO_REG)
   (set (match_operand:HI 0 "register_operand" "=r") (reg:HI REG_PC))]
  ""
  "mov\tr0, %0"
  [(set_attr "length" "2")])
  
;; =================================================================================================

(define_peephole2
  [(set (match_operand:HI 0 "register_operand" "") 
	(match_operand:HI 1 "general_operand" ""))
   (set (mem:HI (post_dec (reg:HI 1)))
	(match_dup 0))]
"dead_or_set_in_peep(1, insn, operands[0])"
  [(set (mem:HI (post_dec (reg:HI 1)))
	(match_dup 1))]
"")

(define_peephole2
  [(set (match_operand:SI 0 "register_operand" "") 
	(match_operand:SI 1 "general_operand" ""))
   (set (mem:SI (post_dec (reg:HI 1)))
	(match_dup 0))]
"dead_or_set_in_peep(1, insn, operands[0])"
  [(set (mem:SI (post_dec (reg:HI 1)))
	(match_dup 1))]
"")

(define_peephole2
  [(set (match_operand:SF 0 "register_operand" "") 
	(match_operand:SF 1 "general_operand" ""))
   (set (mem:SF (post_dec (reg:HI 1)))
	(match_dup 0))]
"dead_or_set_in_peep(1, insn, operands[0])"
  [(set (mem:SF (post_dec (reg:HI 1)))
	(match_dup 1))]
"")


;; ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
;; This instructin sets Z flag

(define_insn "sez"
  [(set (cc0) (const_int 0))]
""
"setz"
  [(set_attr "length" "1")
   (set_attr "cc" "compare")])



;;========================================================================
;; compare

(define_expand "cmpqi"
  [(set (cc0)
        (compare:QI (match_operand:QI 0 "nonimmediate_operand_msp430" "rm")   
                    (match_operand:QI 1 "general_operand_msp430" "rmi")))]
 ""
"
  msp430_compare_op0 = operands[0];
  msp430_compare_op1 = operands[1];
  DONE;
")


(define_expand "cmphi"
  [(set (cc0)
        (compare:HI (match_operand:HI 0 "nonimmediate_operand_msp430" "rm")   
                    (match_operand:HI 1 "general_operand_msp430" "rmi")))]
 ""
"
  msp430_compare_op0 = operands[0];
  msp430_compare_op1 = operands[1];
  DONE;
")


(define_expand "cmpsi"
   [(set (cc0)
         (compare:SI (match_operand:SI 0 "nonimmediate_operand" "rm")
                     (match_operand:SI 1 "general_operand" "rmi")))]
   ""
"
  msp430_compare_op0 = operands[0];
  msp430_compare_op1 = operands[1];
  DONE;
")


(define_expand "beq"
   [(use (match_operand 0 "" ""))]
""
"{ msp430_emit_cbranch (EQ, operands[0]); DONE; }")

(define_expand "bne"
   [(use (match_operand 0 "" ""))]
""
"{ msp430_emit_cbranch (NE, operands[0]); DONE; }")

(define_expand "bge"
   [(use (match_operand 0 "" ""))]
""
"{ msp430_emit_cbranch (GE, operands[0]); DONE; }")

(define_expand "bgt"
   [(use (match_operand 0 "" ""))]
""
"{ msp430_emit_cbranch (GT, operands[0]); DONE; }")

(define_expand "ble"
   [(use (match_operand 0 "" ""))]
""
"{ msp430_emit_cbranch (LE, operands[0]); DONE; }")

(define_expand "blt"
   [(use (match_operand 0 "" ""))]
""
"{ msp430_emit_cbranch (LT, operands[0]); DONE; }")

(define_expand "bgeu"
   [(use (match_operand 0 "" ""))]
""
"{ msp430_emit_cbranch (GEU, operands[0]); DONE; }")

(define_expand "bgtu"
   [(use (match_operand 0 "" ""))]
""
"{ msp430_emit_cbranch (GTU, operands[0]); DONE; }")

(define_expand "bleu"
   [(use (match_operand 0 "" ""))]
""
"{ msp430_emit_cbranch (LEU, operands[0]); DONE; }")

(define_expand "bltu"
   [(use (match_operand 0 "" ""))]
""
"{ msp430_emit_cbranch (LTU, operands[0]); DONE; }")


(define_insn "*cbranchqi"
  [(set (pc)
     (if_then_else (match_operator:QI 1 "comparison_operator"
		    [(match_operand:QI 2 "nonimmediate_operand_msp430" "rm")
		     (match_operand:QI 3 "general_operand" "rmi")])
      (label_ref (match_operand 0 "" ""))
      (pc)))]
""
"* return msp430_cbranch(insn, operands, NULL, 0);"
 [(set_attr "length" "9") 
 (set_attr "cc" "cbranch")])



(define_insn "*cbranchhi"
  [(set (pc)
     (if_then_else (match_operator:HI 1 "comparison_operator"
		    [(match_operand:HI 2 "nonimmediate_operand_msp430" "rm")
		     (match_operand:HI 3 "general_operand" "rmi")])
      (label_ref (match_operand 0 "" ""))
      (pc)))]
""
"* return msp430_cbranch(insn, operands, NULL, 0);"
 [(set_attr "length" "9")
 (set_attr "cc" "cbranch")])


(define_insn "*cbranchsi_eqne"
  [(set (pc)
     (if_then_else (match_operator:SI 1 "equality_operator"
		    [(match_operand:SI 2 "nonimmediate_operand" "rm")
		     (match_operand:SI 3 "general_operand" "rmi")])
      (label_ref (match_operand 0 "" ""))
      (pc)))]
""
"* return msp430_cbranch(insn, operands, NULL, 0);"
[(set_attr "length" "9")
 (set_attr "cc" "cbranch")])


(define_insn "*cbranchsi_others"
  [(parallel [(set (pc)
     (if_then_else (match_operator:SI 1 "inequality_operator"
		    [(match_operand:SI 2 "register_operand" "r")
		     (match_operand:SI 3 "general_operand" "rmi")])
      (label_ref (match_operand 0 "" ""))
      (pc)))
  (clobber (match_dup 2))])]
""
"* return msp430_cbranch(insn, operands, NULL, 0);"
[(set_attr "length" "9")
 (set_attr "cc" "cbranch")])


(define_insn "*cbranch_uncoded"
  [(set (pc)
     (if_then_else (match_operator 1 "comparison_operator"
		    [(cc0)
		     (const_int 0)])
      (label_ref (match_operand 0 "" ""))
      (pc)))]
""
"* return msp430_cbranch(insn, operands, NULL, 1);"
[(set_attr "length" "9")
(set_attr "cc" "cbranch")])


;;========================================================================
;; noop
(define_insn "nop"
  [(const_int 0)]
  ""
  "nop"
  [(set_attr "cc" "none")
   (set_attr "length" "1")])


;;============================================================================
;; call
;;

(define_expand "call"
  [(call (match_operand:HI 0 "general_operand" "")
         (match_operand:HI 1 "general_operand" ""))]
  ""
  "")

(define_insn "*call_insn"
  [(call (mem:HI (match_operand:HI 0 "general_operand" "r,P,mi"))
         (match_operand:HI 1 "general_operand" "X,X,X"))]
""
"* return msp430_emit_call(operands);"
[ (set_attr "length" "1,1,2")
   (set_attr "cc" "clobber")])


(define_expand "call_value"
  [(set (match_operand 0 "register_operand" "")
        (call (match_operand:HI 1 "general_operand" "")
              (match_operand:HI 2 "general_operand" "")))]
  ""
  "")

(define_insn "*call_value_insn"
  [( set (match_operand 0 "register_operand" "=r,r,r")
   (call (mem:HI (match_operand:HI 1 "general_operand" "r,P,mi"))
	 (match_operand:HI 2 "general_operand" "X,X,X")))]
""
 "call\\t%N1"
[ (set_attr "length" "1,1,2")
   (set_attr "cc" "clobber")])





;;========================================================================
;;========================================================================
;; mult helpers

(define_insn "reent_in"
  [(set (mem:HI (post_dec (reg:HI 1))) 
        (unspec_volatile:HI [(const_int 99999999)] 10))]
  ""
  "push\\tr2
\\tdint
\\tnop"
  [(set_attr "length" "4")
   (set_attr "cc" "clobber")])

;;
;; Next three help to make sure, that
;; all instructions are 'in order'

(define_insn "fetch_result_qi"
 [(set (match_operand:QI 0 "nonimmediate_operand_msp430" "=r,m") 
       (unspec_volatile:QI [(const_int 0)] 12))
  (set (mem:HI (post_inc (reg:HI 1)))
	(unspec_volatile:HI [(const_int 99999999)] 11))]
 ""
 "mov.b\\t&__RESLO, %0
\\tpop\\tr2"
 [(set_attr "length" "3,4")
  (set_attr "cc" "none")])
 
(define_insn "fetch_result_hi"
 [(set (match_operand:HI 0 "nonimmediate_operand_msp430" "=r,m") 
       (unspec_volatile:HI [(const_int 0)] 13))
  (set (mem:HI (post_inc (reg:HI 1)))
	(unspec_volatile:HI [(const_int 99999999)] 11))]
 ""
 "mov\\t&__RESLO, %0
\\tpop\\tr2"
 [(set_attr "length" "3,4")
  (set_attr "cc" "none")])

(define_insn "fetch_result_si"
 [(set (match_operand:SI 0 "nonimmediate_operand_msp430" "=r,m") 
       (unspec_volatile:SI [(const_int 0)] 14))
  (set (mem:HI (post_inc (reg:HI 1)))
	(unspec_volatile:HI [(const_int 99999999)] 11))]
 ""
 "mov\\t&__RESLO, %A0
\\tmov\\t&__RESHI, %B0
\\tpop\\tr2"
 [(set_attr "length" "5,7")
  (set_attr "cc" "none")])

;; ===the same with no int

(define_insn "fetch_result_qi_nint"
 [(set (match_operand:QI 0 "nonimmediate_operand_msp430" "=r,m") 
       (unspec_volatile:QI [(const_int 0)] 121))]
 ""
 "mov.b\\t&__RESLO, %0"
 [(set_attr "length" "2,3")
  (set_attr "cc" "none")])
 
(define_insn "fetch_result_hi_nint"
 [(set (match_operand:HI 0 "nonimmediate_operand_msp430" "=r,m") 
       (unspec_volatile:HI [(const_int 0)] 131))]
 ""
 "mov\\t&__RESLO, %0"
 [(set_attr "length" "2,3")
  (set_attr "cc" "none")])

(define_insn "fetch_result_si_nint"
 [(set (match_operand:SI 0 "nonimmediate_operand_msp430" "=r,m") 
       (unspec_volatile:SI [(const_int 0)] 141))]
 ""
 "mov\\t&__RESLO, %A0
\\tmov\\t&__RESHI, %B0"
 [(set_attr "length" "4,6")
  (set_attr "cc" "none")])

(define_insn "addc_zero"
 [(set (match_operand:HI 0 "nonimmediate_operand_msp430" "=r,m")
   (unspec_volatile:HI [(const_int 0 )] 15))]
 ""
 "addc\\t#0, %0"
[(set_attr "length" "1,2")
 (set_attr "cc" "none")])

(define_insn "subc_zero"
 [(set (match_operand:HI 0 "nonimmediate_operand_msp430" "=r,m")
   (unspec:HI [(const_int 0 )] 16))]
 ""
 "subc\\t#0, %0"
[(set_attr "length" "1,2")
 (set_attr "cc" "none")])

(define_insn "load_mpy"
 [(unspec_volatile:HI [(const_int 0)] 17)
  (use (match_operand:HI 0 "general_operand_msp430" "rRP,mi"))]
""
"mov\\t%0, &__MPY"
[(set_attr "length" "2,3")
 (set_attr "cc" "none")])


(define_insn "load_mpys"
  [(unspec_volatile:HI [(const_int 0)] 18)
   (use (match_operand:HI 0 "general_operand_msp430" "rRP,mi"))]
""
"mov\\t%0, &__MPYS"
[(set_attr "length" "2,3")
 (set_attr "cc" "none")])
 
 
(define_insn "load_op2"
  [(unspec_volatile:HI [(const_int 0)] 19)
   (use (match_operand:HI 0 "general_operand_msp430" "rRP,mi"))]
""
"mov\\t%0, &__OP2"
[(set_attr "length" "2,3")
 (set_attr "cc" "none")])
 

(define_insn "load_mpyq"
  [(unspec_volatile:QI [(const_int 0 )] 20)
   (use (match_operand:QI 0 "general_operand_msp430" "rRP,mi"))]
""
"mov.b\\t%0, &__MPY"
[(set_attr "length" "2,3")
 (set_attr "cc" "none")])


(define_insn "load_mpysq"
  [(unspec_volatile:QI [(const_int 0 )] 21)
   (use (match_operand:QI 0 "general_operand_msp430" "rRP,mi"))]
""
"mov.b\\t%0, &__MPYS"
[(set_attr "length" "2,3")
 (set_attr "cc" "none")])
 
(define_insn "load_op2q"
  [(unspec_volatile:QI [(const_int 0 )] 22)
  (use (match_operand:QI 0 "general_operand_msp430" "rRP,mi"))]
""
"mov.b\\t%0, &__OP2"
[(set_attr "length" "2,3")
 (set_attr "cc" "none")])




 

;;========================================================================
;;========================================================================
;;========================================================================
;;
;;  Multiplication 

;;========================================================================
;; 8 = 8x8 and 16 = 8x8

(define_expand "mulqi3"
  [(set (match_operand:QI 0 "nonimmediate_operand_msp430" "")
        (mult:QI (match_operand:QI 1 "nonimmediate_operand_msp430" "")
                 (match_operand:QI 2 "general_operand_msp430" "")))]  
  ""             
"{ msp430_mul3_guard(operands,0); DONE; }")

(define_insn "*mulqi3_call"
  [(set (reg:QI 14) (mult:QI (reg:QI 10) (reg:QI 12)))
	(clobber (reg:QI 10))
	(clobber (reg:QI 12))]
  "!MSP430_HAS_HWMUL_INTERNAL"
  "call	#__mulqi3"
  [(set_attr "length" "2")
   (set_attr "cc" "clobber")])

;; ============= qi -> hi =======================================================
(define_expand "mulqihi3"
  [(set (match_operand:HI 0 "nonimmediate_operand_msp430" "")
        (mult:HI (sign_extend:HI (match_operand:QI 1 "nonimmediate_operand_msp430" ""))
                 (sign_extend:HI (match_operand:QI 2 "general_operand_msp430" ""))))]
""
"{ msp430_mul3_guard(operands,1); DONE; }")

(define_insn "*mulqihi3_call"
  [(set (reg:HI 14) (mult:HI (sign_extend:HI (reg:QI 10))
			     (sign_extend:HI (reg:QI 12))))
        (clobber (reg:QI 10))
        (clobber (reg:QI 12))]
  "!MSP430_HAS_HWMUL_INTERNAL"
  "call	#__mulqihi3"   
  [(set_attr "length" "2")
   (set_attr "cc" "clobber")])

;; ============ unsigned ones ===================================================
(define_expand "umulqihi3"
  [(set (match_operand:HI 0 "nonimmediate_operand_msp430" "")
        (mult:HI (zero_extend:HI (match_operand:QI 1 "nonimmediate_operand_msp430" ""))
                 (zero_extend:HI (match_operand:QI 2 "general_operand_msp430" ""))))]
  ""
  "{ msp430_umul3_guard(operands,0); DONE; }")  

(define_insn "*umulqihi3_call"
  [(set (reg:HI 14) (mult:HI (zero_extend:HI (reg:QI 10))
                	     (zero_extend:HI (reg:QI 12))))
        (clobber (reg:QI 10))
        (clobber (reg:QI 12))]
  "!MSP430_HAS_HWMUL_INTERNAL"
  "call	#__umulqihi3" 
  [(set_attr "length" "2")
   (set_attr "cc" "clobber")])

;;========================================================================
;; 16 = 16x16 and 32 = 16x16

(define_expand "mulhi3"
  [(set (match_operand:HI 0 "nonimmediate_operand_msp430" "")
        (mult:HI (match_operand:HI 1 "nonimmediate_operand_msp430" "")
                 (match_operand:HI 2 "general_operand_msp430" "")))]  
  ""
"{msp430_mul3_guard(operands,0); DONE; } ")

(define_insn "*mulhi3_call"
  [(set (reg:HI 14) (mult:HI (reg:HI 10) (reg:HI 12)))
	(clobber (reg:HI 10))
	(clobber (reg:HI 12))]
  "!MSP430_HAS_HWMUL_INTERNAL"
  "call	#__mulhi3"
  [(set_attr "length" "2")
   (set_attr "cc" "clobber")])

;; ========================== hi -> si =============================
(define_expand "mulhisi3"
  [(set (match_operand:SI 0 "nonimmediate_operand_msp430" "")
        (mult:SI (sign_extend:SI (match_operand:HI 1 "nonimmediate_operand_msp430" ""))
                 (sign_extend:SI (match_operand:HI 2 "general_operand_msp430" ""))))]
""
  "{msp430_mulhisi_guard(operands); DONE;}"
)

(define_insn "*mulhisi3_call"
  [(set (reg:SI 14) (mult:SI (sign_extend:SI (reg:HI 10))
			     (sign_extend:SI (reg:HI 12))))
        (clobber (reg:HI 10))
        (clobber (reg:HI 11))
	(clobber (reg:HI 12))  
	(clobber (reg:HI 13))]
  "!MSP430_HAS_HWMUL_INTERNAL"
  "mov	#0, r11
	tst	r10
	jge	+2
	mov	#-1, r11
	mov	#0, r13
	tst	r12
	jge	+2
	mov	#-1, r13
	call	#__mulhisi3"
  [(set_attr "length" "10")
   (set_attr "cc" "clobber")])

;; ================== unsigned  hi -> si =============================
(define_expand "umulhisi3"
  [(set (match_operand:SI 0 "nonimmediate_operand_msp430" "")
        (mult:SI (zero_extend:SI (match_operand:HI 1 "nonimmediate_operand_msp430" ""))
                 (zero_extend:SI (match_operand:HI 2 "general_operand_msp430" ""))))]
  ""
  "{msp430_umulhisi_guard(operands); DONE;}")

(define_insn "*umulhisi3_call"
  [(set (reg:SI 14) (mult:SI (zero_extend:SI (reg:HI 10))
                	     (zero_extend:SI (reg:HI 12))))
        (clobber (reg:HI 10))
        (clobber (reg:HI 11))
	(clobber (reg:HI 12))  
	(clobber (reg:HI 13))]
  "!MSP430_HAS_HWMUL_INTERNAL"
  "clr	r11
	clr	r13
	call	#__umulhisi3" 
  [(set_attr "length" "2")
   (set_attr "cc" "clobber")])


;;========================================================================
;; 32 = 32x32.      64 = 32x32 <- via library calls only

(define_expand "mulsi3"
  [(set (reg:SI 10) (match_operand:SI 1 "nonimmediate_operand_msp430" ""))
   (set (reg:SI 12) (match_operand:SI 2 "general_operand_msp430" ""))
   (set (reg:SI 14) (mult:SI (reg:SI 10) (reg:SI 12)))
   (set (match_operand:SI 0 "nonimmediate_operand_msp430" "") (reg:SI 14))]
""
"{
  if (!MSP430_HAS_HWMUL_INTERNAL)
    {
      emit_insn (gen_mulsi3_call (operands[0], operands[1], operands[2]));
      DONE;
    }
}")

(define_insn "*mulsi3hw_call_ni"
  [(set (reg:SI 14) (mult:SI (reg:SI 10) (reg:SI 12)))]
  "(!TARGET_INLINESIHWMUL) && MSP430_NOINT_HWMUL"
  "call	#__umulsi3hw"
[(set_attr "length" "2")
   (set_attr "cc" "clobber")])

(define_insn "*mulsi3hw_call_ie"
  [(set (reg:SI 14) (mult:SI (reg:SI 10) (reg:SI 12)))]
  "(!TARGET_INLINESIHWMUL) && !MSP430_NOINT_HWMUL"
  "push	r2
	dint
	call	#__umulsi3hw
	pop	r2"
[(set_attr "length" "5")
   (set_attr "cc" "clobber")])

(define_insn "*mulsi3hw_inline_ni"
  [(set (reg:SI 14) (mult:SI (reg:SI 10) (reg:SI 12)))]
  "TARGET_INLINESIHWMUL && MSP430_HAS_HWMUL_INTERNAL && MSP430_NOINT_HWMUL"
  "mov	r12, &__MPY
	mov	r10, &__OP2
	mov	r12, &__MAC
	mov	&__RESLO, r14
	mov	&__RESHI, &__RESLO
	mov	r11, &__OP2
	mov	r13, &__MAC
	mov	r10, &__OP2
	mov	&__RESLO, r15"
[(set_attr "length" "19")  
   (set_attr "cc" "none")])

(define_insn "*mulsi3hw_inline_ie"
  [(set (reg:SI 14) (mult:SI (reg:SI 10) (reg:SI 12)))]
  "TARGET_INLINESIHWMUL && MSP430_HAS_HWMUL_INTERNAL && !MSP430_NOINT_HWMUL"
  "push	r2
	dint
	nop
	mov	r12, &__MPY
	mov	r10, &__OP2
	mov	r12, &__MAC
	mov	&__RESLO, r14
	mov	&__RESHI, &__RESLO
	mov	r11, &__OP2
	mov	r13, &__MAC
	mov	r10, &__OP2
	mov	&__RESLO, r15
	pop	r2"
[(set_attr "length" "23")  
   (set_attr "cc" "none")])

(define_expand "mulsi3_call"
  [(set (reg:SI 10) (match_operand:SI 1 "register_operand" ""))
   (set (reg:SI 12) (match_operand:SI 2 "register_operand" "")) 
   (parallel [(set (reg:SI 14) (mult:SI (reg:SI 10) (reg:SI 12)))
        (clobber (reg:SI 10))
        (clobber (reg:SI 12))])
   (set (match_operand:SI 0 "register_operand" "") (reg:SI 14))]
"!MSP430_HAS_HWMUL_INTERNAL"
"")

(define_insn "*mulsi3_call"
  [(set (reg:SI 14) (mult:SI (reg:SI 10) (reg:SI 12)))
	(clobber (reg:SI 10))
	(clobber (reg:SI 12))]
  "!MSP430_HAS_HWMUL_INTERNAL"
  "call	#__mulsi3"
  [(set_attr "length" "2")
   (set_attr "cc" "clobber")])


;; / % / % / % / % / % / % / % / % / % / % / % / % / % / % / % / % / % / % /
;; / % / % / % / % / % / % / % / % / % / % / % / % / % / % / % / % / % / % /
;; / % / % / % / % / % / % / % / % / % / % / % / % / % / % / % / % / % / % /
;; / % / % / % / % / % / % / % / % / % / % / % / % / % / % / % / % / % / % /

(define_expand "divmodqi4"
  [(set (reg:QI 12) (match_operand:QI 1 "register_operand" ""))
   (set (reg:QI 10) (match_operand:QI 2 "register_operand" ""))
   (parallel [(set (reg:QI 12) (div:QI (reg:QI 12) (reg:QI 10)))
              (set (reg:QI 14) (mod:QI (reg:QI 12) (reg:QI 10)))
              (clobber (reg:QI 10))
              (clobber (reg:QI 11))
	      (clobber (reg:QI 13))])
   (set (match_operand:QI 0 "register_operand" "") (reg:QI 12))
   (set (match_operand:QI 3 "register_operand" "") (reg:QI 14))]
  ""
  "")

(define_insn "*divmodqi4_call"
  [(set (reg:QI 12) (div:QI (reg:QI 12) (reg:QI 10)))
   (set (reg:QI 14) (mod:QI (reg:QI 12) (reg:QI 10)))
   (clobber (reg:QI 10))
   (clobber (reg:QI 11))
   (clobber (reg:QI 13))]
  ""
  "call	#__divmodqi4"
   [(set_attr "length" "2")
   (set_attr "cc" "clobber")])

(define_expand "udivmodqi4"
  [(set (reg:QI 12) (match_operand:QI 1 "register_operand" ""))
   (set (reg:QI 10) (match_operand:QI 2 "register_operand" ""))
   (parallel [(set (reg:QI 12) (udiv:QI (reg:QI 12) (reg:QI 10)))
              (set (reg:QI 14) (umod:QI (reg:QI 12) (reg:QI 10)))
              (clobber (reg:QI 10))
              (clobber (reg:QI 11)) 
              (clobber (reg:QI 13))])
   (set (match_operand:QI 0 "register_operand" "") (reg:QI 12))
   (set (match_operand:QI 3 "register_operand" "") (reg:QI 14))]
  ""
  "")

(define_insn "*udivmodqi4_call"
  [(set (reg:QI 12) (udiv:QI (reg:QI 12) (reg:QI 10)))
   (set (reg:QI 14) (umod:QI (reg:QI 12) (reg:QI 10)))
   (clobber (reg:QI 10))
   (clobber (reg:QI 11)) 
   (clobber (reg:QI 13))]
  ""
  "call	#__udivmodqi4"
   [(set_attr "length" "2")
   (set_attr "cc" "clobber")])


(define_expand "divmodhi4"
  [(set (reg:HI 12) (match_operand:HI 1 "register_operand" ""))
   (set (reg:HI 10) (match_operand:HI 2 "register_operand" ""))
   (parallel [(set (reg:HI 12) (div:HI (reg:HI 12) (reg:HI 10)))
              (set (reg:HI 14) (mod:HI (reg:HI 12) (reg:HI 10)))
              (clobber (reg:HI 10))
              (clobber (reg:HI 11))
	      (clobber (reg:HI 13))])
   (set (match_operand:HI 0 "register_operand" "") (reg:HI 12))
   (set (match_operand:HI 3 "register_operand" "") (reg:HI 14))]
  ""
  "")

(define_insn "*divmodhi4_call"
  [(set (reg:HI 12) (div:HI (reg:HI 12) (reg:HI 10)))
   (set (reg:HI 14) (mod:HI (reg:HI 12) (reg:HI 10)))
   (clobber (reg:HI 10))
   (clobber (reg:HI 11))
   (clobber (reg:HI 13))]
  ""
  "call	#__divmodhi4"
   [(set_attr "length" "2")
   (set_attr "cc" "clobber")])

(define_expand "udivmodhi4"
  [(set (reg:HI 12) (match_operand:HI 1 "register_operand" ""))
   (set (reg:HI 10) (match_operand:HI 2 "register_operand" ""))
   (parallel [(set (reg:HI 12) (udiv:HI (reg:HI 12) (reg:HI 10)))
              (set (reg:HI 14) (umod:HI (reg:HI 12) (reg:HI 10)))
              (clobber (reg:HI 10))
              (clobber (reg:HI 11)) 
              (clobber (reg:HI 13))])
   (set (match_operand:HI 0 "register_operand" "") (reg:HI 12))
   (set (match_operand:HI 3 "register_operand" "") (reg:HI 14))]
  ""
  "")

(define_insn "*udivmodhi4_call"
  [(set (reg:HI 12) (udiv:HI (reg:HI 12) (reg:HI 10)))
   (set (reg:HI 14) (umod:HI (reg:HI 12) (reg:HI 10)))
   (clobber (reg:HI 10))
   (clobber (reg:HI 11)) 
   (clobber (reg:HI 13))]
  ""
  "call	#__udivmodhi4"
   [(set_attr "length" "2")
   (set_attr "cc" "clobber")])


;; ///////////////// SINGLE INTEGER %%%%%%%%%%%%%%%%%

(define_expand "divmodsi4"
  [(set (reg:SI 12) (match_operand:SI 1 "register_operand" ""))
   (set (reg:SI 10) (match_operand:SI 2 "register_operand" ""))
   (parallel [(set (reg:SI 12) (div:SI (reg:SI 12) (reg:SI 10)))
              (set (reg:SI 14) (mod:SI (reg:SI 12) (reg:SI 10)))
              (clobber (reg:SI 10))
              (clobber (reg:HI 9))
	      (clobber (reg:HI 8))])
   (set (match_operand:SI 0 "register_operand" "") (reg:SI 12))
   (set (match_operand:SI 3 "register_operand" "") (reg:SI 14))]
  ""
  "")

(define_insn "*divmodsi4_call"
  [(set (reg:SI 12) (div:SI (reg:SI 12) (reg:SI 10)))
   (set (reg:SI 14) (mod:SI (reg:SI 12) (reg:SI 10)))
   (clobber (reg:SI 10))
   (clobber (reg:HI 9))
   (clobber (reg:HI 8))]
  ""
  "call	#__divmodsi4"
   [(set_attr "length" "2")
   (set_attr "cc" "clobber")])

(define_expand "udivmodsi4"
  [(set (reg:SI 12) (match_operand:SI 1 "register_operand" ""))
   (set (reg:SI 10) (match_operand:SI 2 "register_operand" ""))
   (parallel [(set (reg:SI 12) (udiv:SI (reg:SI 12) (reg:SI 10)))
              (set (reg:SI 14) (umod:SI (reg:SI 12) (reg:SI 10)))
              (clobber (reg:SI 10))
              (clobber (reg:HI 9)) 
              (clobber (reg:HI 8))])
   (set (match_operand:SI 0 "register_operand" "") (reg:SI 12))
   (set (match_operand:SI 3 "register_operand" "") (reg:SI 14))]
  ""
  "")

(define_insn "*udivmodsi4_call"
  [(set (reg:SI 12) (udiv:SI (reg:SI 12) (reg:SI 10)))
   (set (reg:SI 14) (umod:SI (reg:SI 12) (reg:SI 10)))
   (clobber (reg:SI 10))
   (clobber (reg:HI 9)) 
   (clobber (reg:HI 8))]
  ""
  "call	#__udivmodsi4"
   [(set_attr "length" "2")
   (set_attr "cc" "clobber")])




;;========================================================================
;; MOV STRING
;;   structures and stuff are word aligned.
;;   so, QI mode only defined (as HI actually)
;;

(define_expand "movstrhi"
  [(parallel [(set (match_operand:BLK 0 "memory_operand" "")
                   (match_operand:BLK 1 "memory_operand" ""))
              (use (match_operand 2 "const_int_operand" ""))
              (use (match_operand 3 "const_int_operand" ""))
              (clobber (match_dup 4))
              (clobber (match_dup 5))
              (clobber (match_dup 6))])]
  ""
  " 
{   
  rtx addr0, addr1;
  rtx a0, a1;
  
  if (GET_CODE (operands[2]) != CONST_INT) FAIL;

  addr0 = copy_to_mode_reg (Pmode, XEXP (operands[0], 0));
  addr1 = copy_to_mode_reg (Pmode, XEXP (operands[1], 0));
  
  a0 = operands[0];
  a1 = operands[1];
  
  operands[5] = addr0;
  operands[6] = addr1;

  operands[0] = gen_rtx_MEM (BLKmode, addr0);
  operands[1] = gen_rtx_MEM (BLKmode, addr1);

  if(INTVAL (operands[2]) <= 10 && !(INTVAL(operands[3])&1))
  {
    int x = INTVAL (operands[2]);
    int y = (x&~1) >> 1;
    int i = 0;
     
    while(y--)
    {
      rtx dest = gen_rtx_MEM (HImode, gen_rtx_PLUS(HImode, addr0,GEN_INT(i)));
      emit_insn(gen_movstrhi5(dest,addr1));
      i+= 2;
    }

    if(x & 1)
    {
      rtx real_dst = gen_rtx_MEM (HImode, gen_rtx_PLUS(HImode, addr0,GEN_INT(x-1)));
      emit_insn(gen_movstrqi5(real_dst,addr1));
    }
    DONE;
  }
  else if(INTVAL (operands[2]) <= 6 && (INTVAL(operands[3])&1))
  {
    int x = INTVAL (operands[2]);
    int i = 0;
    
    while(x--)
    {
      rtx dst = gen_rtx_MEM (HImode, gen_rtx_PLUS(HImode, addr0,GEN_INT(i)));
      emit_insn(gen_movstrqi5(dst,addr1));
      i++;
    }
    DONE;
  }
  else
  {
  	operands[2] = copy_to_mode_reg (HImode, operands[2]);
  	operands[4] = operands[2];
  }
}
")

(define_insn "movstrqi5"
  [(set (match_operand:HI 0 "nonimmediate_operand_msp430" "=m")
	(mem:HI (match_operand:HI 1 "register_operand" "+r")))
	(set (match_dup 1) (plus:HI (match_dup 1) (const_int 1)))]
  ""
  "mov.b	@%1+, %0"
[(set_attr "length" "2")
 (set_attr "cc" "clobber")])

(define_insn "movstrhi5"
  [(set (match_operand:HI 0 "nonimmediate_operand_msp430" "=m")
	(mem:HI (match_operand:HI 1 "register_operand" "+r")))
	(set (match_dup 1) (plus:HI (match_dup 1) (const_int 2)))]
 ""
 "mov	@%1+, %0"
 [(set_attr "length" "2")
 (set_attr "cc" "clobber")])

(define_insn "*movstrhi_insn"
  [(set (mem:BLK (match_operand:HI 0 "register_operand" "r"))
        (mem:BLK (match_operand:HI 1 "register_operand" "r")))
   (use (match_operand:HI 2 "register_operand" "r"))
   (use (match_operand 3 "const_int_operand" "i"))
   (clobber (match_dup 2))
   (clobber (match_dup 0))
   (clobber (match_dup 1))]
  ""
  "* return movstrhi_insn(insn, operands, NULL);"
  [(set_attr "length" "6")
   (set_attr "cc" "clobber")])


(define_insn "*movstrqi_insn"
  [(set (mem:BLK (match_operand:HI 0 "register_operand" "r"))
        (mem:BLK (match_operand:HI 1 "register_operand" "r")))
   (use (match_operand:QI 2 "register_operand" "r"))
   (use (match_operand 3 "const_int_operand" "i"))
   (clobber (match_dup 2))
   (clobber (match_dup 0))
   (clobber (match_dup 1))]
  ""
  "* return movstrhi_insn(insn, operands, NULL);"
  [(set_attr "length" "6")
   (set_attr "cc" "clobber")])



;;========================================================================
;; CLEAR STRING

(define_expand "clrstrhi"
  [(parallel [(set (match_operand:BLK 0 "memory_operand" "")
                   (const_int 0))
              (use (match_operand 1 "const_int_operand" ""))
              (use (match_operand 2 "const_int_operand" "i"))
              (clobber (match_dup 3))
              (clobber (match_dup 4))])]
  ""
  " 
{
  rtx addr0;

  if (GET_CODE (operands[1]) != CONST_INT) FAIL;
  operands[1] = copy_to_mode_reg (HImode, operands[1]);
  operands[3] = operands[1];
  addr0 = copy_to_mode_reg (Pmode, XEXP (operands[0], 0));
  operands[4] = addr0;
  operands[0] = gen_rtx_MEM (BLKmode, addr0);
}")


(define_insn "*clrstrhi_insn"
  [(set (mem:BLK (match_operand:HI 0 "register_operand" "r"))
        (const_int 0))
   (use (match_operand:HI 1 "register_operand" "r")) 
   (use (match_operand 2 "const_int_operand" "i"))
   (clobber (match_dup 1)) 
   (clobber (match_dup 0))]
  ""
  "* return clrstrhi_insn(insn, operands, NULL);"
[(set_attr "length" "6")
   (set_attr "cc" "clobber")])


;;========================================================================
;; %0 = strchr(%1,%2) - %1

(define_expand "strlenhi"
  [(set (match_dup 4)
            (unspec:HI [(match_operand:BLK 1 "memory_operand" "")
                        (match_operand 2 "const_int_operand" "")
                        (match_operand:HI 3 "immediate_operand" "")] 3))
     (set (match_operand:HI 0 "register_operand" "")
          (minus:HI (match_dup 4)
                    (match_dup 5)))]

   ""
   "
{
  rtx addr;

  if (! (GET_CODE (operands[2]) == CONST_INT && INTVAL (operands[2]) == 0))
	FAIL;
  addr = copy_to_mode_reg (Pmode, XEXP (operands[1],0));
  operands[1] = gen_rtx_MEM (BLKmode, addr); 
  operands[5] = addr;
  operands[4] = gen_reg_rtx (HImode);

}")


(define_insn "*strlenhi"
 [(set (match_operand:HI 0 "register_operand" "=r")
        (unspec:HI [(mem:BLK (match_operand:HI 1 "register_operand" "0"))
                    (const_int 0)
		    (match_operand:HI 2 "immediate_operand" "i") ] 3))]
  ""
"dec	%0
.L__strlenhi__%=:
	inc	%0
	tst.b	0(%0)
        jne	.L__strlenhi__%="
[(set_attr "length" "5")
   (set_attr "cc" "clobber")])


;;========================================================================
;; MOV code
;;
;;


;;========================================================================
;; move byte
;; nothing much special
;; all addressing modes allowed
;; fits perfectly into a single instruction 

(define_expand "movqi" 
  [(set (match_operand:QI 0 "nonimmediate_operand_msp430" "")
        (match_operand:QI 1 "general_operand" ""))]
  ""
  "")

(define_insn "*movqi3"
  [(set (match_operand:QI 0 "nonimmediate_operand"	"=m,m,m,m,r,r,r,r")
	(match_operand:QI 1 "general_operand_msp430"	" m,r,P,i,m,r,P,i"))]
  ""
  "mov.b\\t%1, %0"
  [(set_attr "length" "3,2,3,3,2,1,2,2")
   (set_attr "cc" "none,none,none,none,none,none,none,none")])


(define_insn "movqipi"
  [(set (match_operand:QI 0 "nonimmediate_operand_msp430" "=r,m")
	(mem:QI (post_inc:QI (match_operand:HI 1 "register_operand" "r,r"))))]
  ""
  "mov.b	@%1+, %0"
[(set_attr "length" "1,2")
   (set_attr "cc" "none,none")])



;;============================================================================
;; move word (16 bit)
;; the same as above

(define_expand "movhi" 
  [(set (match_operand:HI 0 "nonimmediate_operand_msp430" "")
        (match_operand:HI 1 "general_operand_msp430" ""))]
  ""
  "msp430_expand_mov_intptr (operands[0], operands[1]); DONE;")

(define_insn "*movhi3"
  [(set (match_operand:HI 0 "nonimmediate_operand_msp430"	"=m,m,m,m,r,r,r,r")
	(match_operand:HI 1 "general_operand_msp430"		" r,m,P,i,r,m,P,i"))]
  ""
  "mov\\t%1, %0 "
  [(set_attr "length" "2,3,3,3,1,2,2,2")
   (set_attr "cc" "none")])

(define_insn "movhipi"
  [(set (match_operand:HI 0 "nonimmediate_operand_msp430" "=r,m")
	(mem:HI (post_inc:HI (match_operand:HI 1 "register_operand" "r,r"))))]
  ""
  "mov	@%1+, %0"
[(set_attr "length" "1,2")
   (set_attr "cc" "none,none")])

;;============================================================================
;; move long (32 bit)
;; the same as above

(define_expand "movsi" 
  [(set (match_operand:SI 0 "nonimmediate_operand" "")
        (match_operand:SI 1 "general_operand" ""))]
  ""
  "")

(define_insn "*movsi3"
  [(set (match_operand:SI 0 "nonimmediate_operand"	"=rm")
	(match_operand:SI 1 "general_operand"		" rmi"))]
""
"* return msp430_movesi_code(insn,operands,NULL);"
  [(set_attr "length" "6")
   (set_attr "cc" "none")])

(define_insn "movsipi"
  [(set (match_operand:SI 0 "nonimmediate_operand_msp430" "=r,m")
	(mem:SI (post_inc (match_operand:HI 1 "register_operand" "r,r"))))]
  ""
  "mov	@%1+, %A0
	mov	@%1+, %B0"
[(set_attr "length" "2,4")
   (set_attr "cc" "none,none")])



;;============================================================================
;; floats are the SI

(define_expand "movsf" 
  [(set (match_operand:SF 0 "nonimmediate_operand" "")
        (match_operand:SF 1 "general_operand" ""))]
  ""
  "")

(define_insn "*movsf3"
  [(set (match_operand:SF 0 "nonimmediate_operand"	"=rm")
	(match_operand:SF 1 "general_operand"		"rmi"))]
""
"* return msp430_movesi_code(insn,operands,NULL);"
  [(set_attr "length" "6")
   (set_attr "cc" "none")])


(define_insn "movsfpi"
  [(set (match_operand:SF 0 "nonimmediate_operand_msp430" "=r,m")
	(mem:SF (post_inc (match_operand:HI 1 "register_operand" "r,r"))))]
  ""
  "mov	@%1+, %A0
	mov	@%1+, %B0"
[(set_attr "length" "2,4")
   (set_attr "cc" "none,none")])

;;============================================================================
;; move long long (64 bit)
;; the same as above
(define_expand "movdi" 
  [(set (match_operand:DI 0 "nonimmediate_operand" "")
        (match_operand:DI 1 "general_operand" ""))]
  ""
  "")

(define_insn "*movdi3"
  [(set (match_operand:DI 0 "nonimmediate_operand" 	"=rm")
	(match_operand:DI 1 "general_operand"		"rmi"))]
""
"* return msp430_movedi_code(insn,operands,NULL);"
  [(set_attr "length" "12")
   (set_attr "cc" "none")])


(define_insn "movdipi"
  [(set (match_operand:DI 0 "nonimmediate_operand_msp430" "=r,m")
	(mem:DI (post_inc (match_operand:HI 1 "register_operand" "r,r"))))]
  ""
  "mov	@%1+, %A0
	mov	@%1+, %B0
	mov	@%1+, %C0
	mov	@%1+, %D0"
[(set_attr "length" "4,8")
   (set_attr "cc" "none,none")])


;;============================================================================
;;  ARITHMETIC CODE
;;


;;  random operations:

(define_insn "*opqi3_pi"
  [(set (match_operand:QI 0 "nonimmediate_operand_msp430"          "=r,m")
	(match_operator:QI 3 "three_operands_msp430"
	 [(match_operand:QI 1 "nonimmediate_operand_msp430" "%0,0")
	 (mem:QI (post_inc (match_operand:HI 2 "register_operand" "r,r")))]))]
""
"%3.b	@%2+, %0"
  [(set_attr "length" "1,2")
   (set_attr "cc" "oper,oper")])

(define_insn "*ophi3_pi"
  [(set (match_operand:HI 0 "nonimmediate_operand_msp430"          "=r,m")
	(match_operator:HI 3 "three_operands_msp430"
	 [(match_operand:HI 1 "nonimmediate_operand_msp430" "%0,0")
	  (mem:HI (post_inc (match_operand:HI 2 "register_operand" "r,r")))]))]
""
"%3	@%2+, %0"
  [(set_attr "length" "1,2")
   (set_attr "cc" "oper,oper")])

(define_insn "*opsi3_pi"
  [(set (match_operand:SI 0 "nonimmediate_operand_msp430"          "=r,m")
	(match_operator:SI 3 "three_operands_msp430"
	 [(match_operand:SI 1 "nonimmediate_operand_msp430" "%0,0")
	  (mem:SI (post_inc (match_operand:HI 2 "register_operand" "r,r")))]))]
""
"%A3\\t@%2+, %A0
\\t%B3\\t@%2+, %B0"
  [(set_attr "length" "1,2")
   (set_attr "cc" "oper,oper")])

(define_insn "*opdi3_pi"
  [(set (match_operand:DI 0 "nonimmediate_operand_msp430"          "=r,m")
	(match_operator:DI 3 "three_operands_msp430"
	 [(match_operand:DI 1 "nonimmediate_operand_msp430" "%0,0")
	  (mem:DI (post_inc (match_operand:HI 2 "register_operand" "r,r")))]))]
""
"%A3\\t@%2+, %A0
\\t%B3\\t@%2+, %B0
\\t%C3\\t@%2+, %C0
\\t%D3\\t@%2+, %D0"
  [(set_attr "length" "1,2")
   (set_attr "cc" "oper,oper")])




;;++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
;;++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
;;++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
;;++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
;;++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
;;++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
;;++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
;; add 1 byte 


(define_expand "addqi3" 
  [(set (match_operand:QI 0 "nonimmediate_operand_msp430" "")
        (plus:QI (match_operand:QI 1 "nonimmediate_operand_msp430" "")   
                 (match_operand:QI 2 "general_operand_msp430" "")))]
  "" 
  "")

(define_insn "*addqi3_cg"
  [(set (match_operand:QI 0 "nonimmediate_operand_msp430" "=m,r")
        (plus:QI (match_operand:QI 1 "nonimmediate_operand_msp430" "%0,0")
		 (match_operand:QI 2 "const_int_operand"      "i,i")))]
"(INTVAL(operands[2]) == -2
  || INTVAL(operands[2]) == -4
  || INTVAL(operands[2]) == -8 )"
"* {
      operands[2] = gen_rtx_CONST_INT(QImode, -INTVAL(operands[2]));
      return \"sub.b\\t%2, %0\";
}"
[(set_attr "length" "2,1")
 (set_attr "cc" "set_czn,set_czn")])


(define_insn "*addqi3_3"
  [(set (match_operand:QI 0 "nonimmediate_operand_msp430"          "=m,m,m,m,r,r,r,r")  
        (plus:QI (match_operand:QI 1 "nonimmediate_operand_msp430" "%0,0,0,0,0,0,0,0")  
                 (match_operand:QI 2 "general_operand_msp430"      " m,r,P,i,m,r,P,i")))]  
""
  "add.b	%2, %0"
  [(set_attr "length" "3,2,2,3,2,1,1,2")
   (set_attr "cc" "set_czn,set_czn,set_czn,set_czn,set_czn,set_czn,set_czn,set_czn")])


;;============================================================================
;; add 1 word (16 bits)
;; same as above


(define_expand "addhi3" 
  [(set (match_operand:HI 0 "nonimmediate_operand_msp430" "")
        (plus:HI (match_operand:HI 1 "nonimmediate_operand_msp430" "")
                 (match_operand:HI 2 "general_operand_msp430" "")))]
  ""
  "")

(define_insn "*addhi3_cg"
  [(set (match_operand:HI 0 "nonimmediate_operand_msp430" "=m,r")
	(plus:HI (match_operand:HI 1 "nonimmediate_operand_msp430" "%0,0")
		 (match_operand:HI 2 "const_int_operand"      "i,i")))]
"(INTVAL(operands[2]) == -2
  || INTVAL(operands[2]) == -4
  || INTVAL(operands[2]) == -8 )"
"* {
      operands[2] = gen_rtx_CONST_INT(HImode, -INTVAL(operands[2]));
      return \"sub\\t%2, %0\" ;
}"
[(set_attr "length" "2,1")
 (set_attr "cc" "set_czn,set_czn")])

(define_insn "*addhi3_3"
  [(set (match_operand:HI 0 "nonimmediate_operand_msp430"          "=m,m,m,m,r,r,r,r")
        (plus:HI (match_operand:HI 1 "nonimmediate_operand_msp430" "%0,0,0,0,0,0,0,0")
                 (match_operand:HI 2 "general_operand_msp430"      " m,r,P,i,m,r,P,i")))]
""
  "add	%2, %0"
  [(set_attr "length" "3,2,2,3,2,1,1,2")
   (set_attr "cc"
"set_czn,set_czn,set_czn,set_czn,set_czn,set_czn,set_czn,set_czn")])


;;============================================================================
;; add 2 words (32 bits)
;; same as above

(define_expand "addsi3"
  [(set (match_operand:SI 0 "nonimmediate_operand_msp430" "")
        (plus:SI (match_operand:SI 1 "nonimmediate_operand_msp430" "")
                 (match_operand:SI 2 "general_operand_msp430" "")))]
  ""
  "")

(define_insn "*addsi3_cg"
  [(set (match_operand:SI 0 "nonimmediate_operand_msp430" "=m,r")
	(plus:SI (match_operand:SI 1 "nonimmediate_operand_msp430" "%0,0")
		 (match_operand:SI 2 "const_int_operand"      "i,i")))]
"(INTVAL(operands[2]) == -2
  || INTVAL(operands[2]) == -4
  || INTVAL(operands[2]) == -8 )"
"* {
     operands[2] = gen_rtx_CONST_INT(SImode, -INTVAL(operands[2]));
     return \"sub\\t%A2, %A0\\n\\tsubc\\t%B2, %B0\" ;
}"
[(set_attr "length" "4,2")
 (set_attr "cc" "further,further")])


(define_insn "*addsi3_3"
  [(set (match_operand:SI 0 "nonimmediate_operand_msp430"      "=rm")
        (plus:SI (match_operand:SI 1 "general_operand_msp430"  "%0")
                 (match_operand:SI 2 "general_operand_msp430"  " rmi")))]
""
"* return msp430_addsi_code(insn, operands, NULL);"
  [(set_attr "length" "6")
   (set_attr "cc" "further")])


;;============================================================================
;; add 4 words (64 bits)
;; same as above

(define_expand "adddi3"
  [(set (match_operand:DI 0 "nonimmediate_operand" "")
        (plus:DI (match_operand:DI 1 "nonimmediate_operand" "")   
                 (match_operand:DI 2 "general_operand" "")))]
  "" 
  "")

(define_insn "*adddi3_cg"
   [(set (match_operand:DI 0 "nonimmediate_operand_msp430" "=m,r")
		  (plus:DI (match_operand:DI 1 "nonimmediate_operand_msp430" "%0,0")
	          (match_operand:DI 2 "const_int_operand"      "i,i")))]
"(INTVAL(operands[2]) == -2
  || INTVAL(operands[2]) == -4
  || INTVAL(operands[2]) == -8 )"
"* {
     operands[2] = gen_rtx_CONST_INT(DImode, -INTVAL(operands[2]));
     return \"sub\\t%A2, %A0\\n\\tsubc\\t%B2, %B0\\n\\tsubc\\t%C2, %C0\\n\\tsubc\\t%D2, %D0\";
}"
[(set_attr "length" "4,2")
(set_attr "cc" "further,further")])

(define_insn "*adddi3_3"
  [(set (match_operand:DI 0 "nonimmediate_operand"      	"=rm")
        (plus:DI (match_operand:DI 1 "nonimmediate_operand"  	"%0")
                 (match_operand:DI 2 "general_operand"  	" rmi")))]
""
"* return msp430_adddi_code(insn, operands, NULL);"
  [(set_attr "length" "12")
   (set_attr "cc" "further")])


;;-----------------------------------------------------------------------
;;-----------------------------------------------------------------------
;;-----------------------------------------------------------------------
;;-----------------------------------------------------------------------
;;-----------------------------------------------------------------------
;; sub 1 byte 


(define_expand "subqi3" 
  [(set (match_operand:QI 0 "nonimmediate_operand_msp430" "")
        (minus:QI (match_operand:QI 1 "nonimmediate_operand_msp430" "")   
                  (match_operand:QI 2 "general_operand_msp430" "")))]
  "" 
  "")

(define_insn "*subqi3_3"
  [(set (match_operand:QI 0 "nonimmediate_operand_msp430"          "=m,m,m,m,r,r,r,r")  
        (minus:QI (match_operand:QI 1 "nonimmediate_operand_msp430" "0,0,0,0,0,0,0,0")  
                 (match_operand:QI 2 "general_operand_msp430"      " m,r,P,i,m,r,P,i")))]  
""
  "sub.b	%2, %0"
  [(set_attr "length" "3,2,2,3,2,1,1,2")
   (set_attr "cc" "set_czn,set_czn,set_czn,set_czn,set_czn,set_czn,set_czn,set_czn")])


;;============================================================================
;; sub 1 word (16 bits)
;; same as above


(define_expand "subhi3" 
  [(set (match_operand:HI 0 "nonimmediate_operand_msp430" "")
        (minus:HI (match_operand:HI 1 "nonimmediate_operand_msp430" "")
                  (match_operand:HI 2 "general_operand_msp430" "")))]
  ""
  "")

(define_insn "*subhi3_3"
  [(set (match_operand:HI 0 "nonimmediate_operand_msp430"          "=m,m,m,m,r,r,r,r")
        (minus:HI (match_operand:HI 1 "nonimmediate_operand_msp430" "0,0,0,0,0,0,0,0")
                  (match_operand:HI 2 "general_operand_msp430"      "m,r,P,i,m,r,P,i")))]
""
  "sub	%2, %0"
  [(set_attr "length" "3,2,2,3,2,1,1,2")
   (set_attr "cc"
"set_czn,set_czn,set_czn,set_czn,set_czn,set_czn,set_czn,set_czn")])


;;============================================================================
;; sub 2 words (32 bits)
;; same as above

(define_expand "subsi3"
  [(set (match_operand:SI 0 "nonimmediate_operand" "")
        (minus:SI (match_operand:SI 1 "nonimmediate_operand" "")
                 (match_operand:SI 2 "general_operand" "")))]
  "" 
  "")


(define_insn "*subsi3_3"
  [(set (match_operand:SI 0 "nonimmediate_operand"      "=rm")
        (minus:SI (match_operand:SI 1 "general_operand"  "0")
                 (match_operand:SI 2 "general_operand"  " rmi")))]
""
"* return msp430_subsi_code(insn, operands, NULL);"
  [(set_attr "length" "6")
   (set_attr "cc" "further")])


;;============================================================================
;; sub 4 words (64 bits)
;; same as above

(define_expand "subdi3"
  [(set (match_operand:DI 0 "nonimmediate_operand" "")
        (minus:DI (match_operand:DI 1 "nonimmediate_operand" "")   
                 (match_operand:DI 2 "general_operand" "")))]
  "" 
  "")

(define_insn "*subdi3_3"
  [(set (match_operand:DI 0 "nonimmediate_operand"      	"=rm")
        (minus:DI (match_operand:DI 1 "nonimmediate_operand"  	"0")
                 (match_operand:DI 2 "general_operand"  	" rmi")))]
""
"* return msp430_subdi_code(insn, operands,NULL);"
  [(set_attr "length" "12")
   (set_attr "cc" "set_n")])





;;&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&
;;&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&
;;&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&
;;&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&
;;&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&
;;&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&
;;&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&
;; and 1 byte 


(define_expand "andqi3" 
  [(set (match_operand:QI 0 "nonimmediate_operand_msp430" "")
        (and:QI (match_operand:QI 1 "nonimmediate_operand_msp430" "")   
                 (match_operand:QI 2 "general_operand_msp430" "")))]
  "" 
  "")


(define_insn "*andqi3_inv"
  [(set (match_operand:QI 0 "nonimmediate_operand_msp430"         "=r,m")   
        (and:QI (match_operand:QI 1 "nonimmediate_operand_msp430" "%0,0")
                (match_operand:QI 2 "immediate_operand"    " i,i")))]
"(INTVAL(operands[2])==~1
  || INTVAL(operands[2])==~2
  || INTVAL(operands[2])==~4
  || INTVAL(operands[2])==~8)"
"* {
	operands[2] = gen_rtx_CONST_INT(QImode, ~(INTVAL(operands[2])));
	return \"bic.b	%2,%0\";
}"
  [(set_attr "length" "1,2")
   (set_attr "cc" "none,none")])


(define_insn "*andqi3_3"
  [(set (match_operand:QI 0 "nonimmediate_operand_msp430"      	"=r,m,r,m,r,m,m,r")
        (and:QI (match_operand:QI 1 "nonimmediate_operand_msp430"  	"%0,0,0,0,0,0,0,0")
                (match_operand:QI 2 "general_operand_msp430"  		" r,m,P,P,m,r,i,i")))]
""
"and.b	%2, %0"
  [(set_attr "length" "1,3,1,2,2,2,3,2")
   (set_attr "cc" "set_czn,set_czn,set_czn,set_czn,set_czn,set_czn,set_czn,set_czn")])




;;============================================================================
;; and 1 word (16 bits)
;; same as above

(define_expand "andhi3" 
  [(set (match_operand:HI 0 "nonimmediate_operand_msp430" "")
        (and:HI (match_operand:HI 1 "nonimmediate_operand_msp430" "%0")   
                 (match_operand:HI 2 "general_operand_msp430" "")))]
  "" 
  "")


(define_insn "*andhi3_inv"
  [(set (match_operand:HI 0 "nonimmediate_operand_msp430"         "=r,m")   
        (and:HI (match_operand:HI 1 "nonimmediate_operand_msp430" "%0,0")   
                (match_operand:HI 2 "immediate_operand"    " i,i")))]
"(INTVAL(operands[2])==~1
  || INTVAL(operands[2])==~2
  || INTVAL(operands[2])==~4
  || INTVAL(operands[2])==~8)"
"* {
	operands[2] = gen_rtx_CONST_INT(HImode, ~(INTVAL(operands[2])));
	return \"bic	%2,%0\";
}"
  [(set_attr "length" "1,2")
   (set_attr "cc" "none,none")])

(define_insn "*andhi3_clrup"
  [(set (match_operand:HI 0 "nonimmediate_operand_msp430"         "=r,m")
        (and:HI (match_operand:HI 1 "nonimmediate_operand_msp430" "%0,0")
                (match_operand:HI 2 "immediate_operand"    " i,i")))]
"INTVAL(operands[2])==255"
"* {
	if(which_alternative == 0)
	{
		return \"and.b	#-1, %0\";
	}
	else if(which_alternative == 1)
	{
		return \"clr.b	%J0\";
	}

	return	\"bug\";
}"
  [(set_attr "length" "1,2")
   (set_attr "cc" "clobber,clobber")])

(define_insn "*andhi3_clrlw"
  [(set (match_operand:HI 0 "nonimmediate_operand_msp430"         "=m,r")
        (and:HI (match_operand:HI 1 "nonimmediate_operand_msp430" "%0,0")
                (match_operand:HI 2 "immediate_operand"     " i,i")))]
"((0xffff&INTVAL(operands[2]))==0xff00)"
"@
clr.b	%I0
and\\t#0xff00, %0"
  [(set_attr "length" "2")
   (set_attr "cc" "clobber")])


(define_insn "*andhi3_3"
  [(set (match_operand:HI 0 "nonimmediate_operand_msp430"      		"=r,m,r,m,r,m,m,r")
        (and:HI (match_operand:HI 1 "nonimmediate_operand_msp430"  	"%0,0,0,0,0,0,0,0")
                (match_operand:HI 2 "general_operand_msp430"  		" r,m,P,P,m,r,i,i")))]
""
"and	%2, %0"
  [(set_attr "length" "1,3,1,2,2,2,3,2")
   (set_attr "cc" "set_czn,set_czn,set_czn,set_czn,set_czn,set_czn,set_czn,set_czn")])





;;============================================================================
;; and 2 words (32 bits)
;; same as above

(define_expand "andsi3"
  [(set (match_operand:SI 0 "nonimmediate_operand" "")
        (and:SI (match_operand:SI 1 "nonimmediate_operand" "")
                (match_operand:SI 2 "general_operand" "")))]
  "" 
  "")

(define_insn "*andsi3_3"
  [(set (match_operand:SI 0 "nonimmediate_operand"          "=rm")
        (and:SI (match_operand:SI 1 "nonimmediate_operand"  "%0")
                (match_operand:SI 2 "general_operand"       " rmi")))]
""
"* return msp430_andsi_code(insn, operands, NULL);"
  [(set_attr "length" "6")
   (set_attr "cc" "set_n")])


;;============================================================================
;; and 4 words (64 bits)
;; same as above

(define_expand "anddi3"
  [(set (match_operand:DI 0 "nonimmediate_operand" "")
        (and:DI (match_operand:DI 1 "nonimmediate_operand" "")   
                (match_operand:DI 2 "general_operand" "")))]
  "" 
  "")

(define_insn "*anddi3_3"
  [(set (match_operand:DI 0 "nonimmediate_operand"         "=rm")
        (and:DI (match_operand:DI 1 "nonimmediate_operand" "%0")
                (match_operand:DI 2 "general_operand"      " rmi")))]
""
"* return msp430_anddi_code(insn, operands, NULL);"
  [(set_attr "length" "14")
   (set_attr "cc" "clobber")])



;;|||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
;;|||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
;;|||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
;;|||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
;;|||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
;;|||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
;;|||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
;; ior 1 byte 
;; looks like a 'mov' insn

(define_expand "iorqi3" 
  [(set (match_operand:QI 0 "nonimmediate_operand_msp430" "")
        (ior:QI (match_operand:QI 1 "nonimmediate_operand_msp430" "")   
                 (match_operand:QI 2 "general_operand_msp430" "")))]
  "" 
  "")

(define_insn "*iorqi3_3"
  [(set (match_operand:QI 0 "nonimmediate_operand_msp430"      	"=r,m,r,m,r,m,m,r")
        (ior:QI (match_operand:QI 1 "nonimmediate_operand_msp430"  	"%0,0,0,0,0,0,0,0")
                (match_operand:QI 2 "general_operand_msp430"  		" r,m,P,P,m,r,i,i")))]
""
  "bis.b	%2, %0"
  [(set_attr "length" "1,3,1,2,2,2,3,2")
   (set_attr "cc" "none,none,none,none,none,none,none,none")])



;;============================================================================
;; ior 1 word (16 bits)
;; same as above

(define_expand "iorhi3" 
  [(set (match_operand:HI 0 "nonimmediate_operand_msp430" "")
        (ior:HI (match_operand:HI 1 "nonimmediate_operand_msp430" "")   
                 (match_operand:HI 2 "general_operand_msp430" "")))]
  "" 
  "
    if(const_int_operand(operands[2], VOIDmode))
    {
      int x = INTVAL(operands[2]) & 0xffff;
      if(!x) DONE;
    }
  ")

(define_insn "*iorhi3_3"
  [(set (match_operand:HI 0 "nonimmediate_operand_msp430"      	"=r,m,r,m,r,m,m,r")
        (ior:HI (match_operand:HI 1 "nonimmediate_operand_msp430"  	"%0,0,0,0,0,0,0,0")
                (match_operand:HI 2 "general_operand_msp430"  		" r,m,P,P,m,r,i,i")))]
""
  "bis	%2, %0"
  [(set_attr "length" "1,3,1,2,2,2,3,2")
   (set_attr "cc" "none,none,none,none,none,none,none,none")])



;;============================================================================
;; ior 2 words (32 bits)
;; same as above


(define_expand "iorsi3"
  [(set (match_operand:SI 0 "nonimmediate_operand" "")
        (ior:SI (match_operand:SI 1 "nonimmediate_operand" "")
                 (match_operand:SI 2 "general_operand" "")))]
  "" 
  "")

(define_insn "*iorsi3_3"
  [(set (match_operand:SI 0 "nonimmediate_operand"      "=rm")
        (ior:SI (match_operand:SI 1 "nonimmediate_operand"   "%0")
                (match_operand:SI 2 "general_operand"   " rmi")))]
""
"* return msp430_iorsi_code(insn, operands, NULL);"
  [(set_attr "length" "6")
   (set_attr "cc" "none")])

(define_split
 [(set (match_operand:SI 0 "nonimmediate_operand"          "")
       (ior:SI (match_operand:SI 1 "nonimmediate_operand"  "")
	       (match_operand:SI 2 "const_int_operand"     "")))]
 "reload_completed
  && (halfnibble_integer(operands[2], VOIDmode)
       || halfnibble_constant(operands[2], VOIDmode))"
 [(set (match_dup 3) (ior:HI (match_dup 4) (match_dup 5)))]
 "{
      int lo = trunc_int_for_mode(INTVAL(operands[2]),HImode);
      int hi = trunc_int_for_mode(INTVAL(operands[2])>>16,HImode);
 
      if(lo == -1)
      {
	rtx op = gen_lowpart(HImode, operands[0]);
	emit_insn(gen_rtx_SET(HImode, op, GEN_INT(-1)));
	DONE;
      }
      
      if(hi == -1)
      {
	rtx op = gen_highpart(HImode, operands[0]);
	emit_insn(gen_rtx_SET(HImode, op, GEN_INT(-1)));
	DONE;
      }
      
      if(lo)
      {
	operands[3] = gen_lowpart(HImode, operands[0]);
	operands[4] = gen_lowpart(HImode, operands[1]);
	operands[5] = GEN_INT(lo);
      }
      else if(hi)
      {
	operands[3] = gen_highpart(HImode, operands[0]);
	operands[4] = gen_highpart(HImode, operands[1]);
	operands[5] = GEN_INT(hi);
      }
 }")

(define_peephole2
 [(set (match_operand:HI 0 "nonimmediate_operand_msp430" "") 
       (const_int 0))
  (set (match_dup 0) (ior:HI (match_dup 0) 
			     (match_operand:HI 1 "const_int_operand" "")))]
 ""
 [(set (match_dup 0) (match_dup 1))]
 "")
 

;;============================================================================
;; ior 4 words (64 bits)
;; same as above

(define_expand "iordi3"
  [(set (match_operand:DI 0 "nonimmediate_operand" "")
        (ior:DI (match_operand:DI 1 "nonimmediate_operand" "")   
                (match_operand:DI 2 "general_operand" "")))]
  "" 
  "")

(define_insn "*iordi3_3"
  [(set (match_operand:DI 0 "nonimmediate_operand"          "=rm")
        (ior:DI (match_operand:DI 1 "nonimmediate_operand"  "%0")
                (match_operand:DI 2 "general_operand"  " rmi")))]
""
"* return msp430_iordi_code(insn, operands, NULL);"
  [(set_attr "length" "12")
   (set_attr "cc" "none")])



;;^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
;;^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
;;^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
;;^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
;;^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
;;^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
;; xor 1 byte 
;; looks like a 'mov' insn

(define_expand "xorqi3" 
  [(set (match_operand:QI 0 "nonimmediate_operand_msp430" "")
        (xor:QI (match_operand:QI 1 "nonimmediate_operand_msp430" "")   
                 (match_operand:QI 2 "general_operand_msp430" "")))]
  "" 
  "")

(define_insn "*xorqi3_3"
  [(set (match_operand:QI 0 "nonimmediate_operand_msp430"      	"=r,m,r,m,r,m,m,r")
        (xor:QI (match_operand:QI 1 "nonimmediate_operand_msp430"  	"%0,0,0,0,0,0,0,0")
                (match_operand:QI 2 "general_operand_msp430"  		" r,m,P,P,m,r,i,i")))]
""
  "xor.b	%2, %0"
  [(set_attr "length" "1,3,1,2,2,2,3,2")
   (set_attr "cc" "set_czn,set_czn,set_czn,set_czn,set_czn,set_czn,set_czn,set_czn")])



;;============================================================================
;; xor 1 word (16 bits)
;; same as above

(define_expand "xorhi3" 
  [(set (match_operand:HI 0 "nonimmediate_operand_msp430" "")
        (xor:HI (match_operand:HI 1 "nonimmediate_operand_msp430" "")   
                 (match_operand:HI 2 "general_operand_msp430" "")))]
  "" 
  "")

(define_insn "*xorhi3_3"
  [(set (match_operand:HI 0 "nonimmediate_operand_msp430"      	"=r,m,r,m,r,m,m,r")
        (xor:HI (match_operand:HI 1 "nonimmediate_operand_msp430"  	"%0,0,0,0,0,0,0,0")
                (match_operand:HI 2 "general_operand_msp430"  		" r,m,P,P,m,r,i,i")))]
""
  "xor	%2, %0"
  [(set_attr "length" "1,3,1,2,2,2,3,2")
   (set_attr "cc" "set_czn,set_czn,set_czn,set_czn,set_czn,set_czn,set_czn,set_czn")])



;;============================================================================
;; xor 2 words (32 bits)
;; same as above

(define_expand "xorsi3"
  [(set (match_operand:SI 0 "nonimmediate_operand" "")
        (xor:SI (match_operand:SI 1 "nonimmediate_operand" "")
                 (match_operand:SI 2 "general_operand" "")))]
  "" 
  "")


(define_insn "*xorsi3_3"
  [(set (match_operand:SI 0 "nonimmediate_operand"      "=rm")
        (xor:SI (match_operand:SI 1 "nonimmediate_operand"   "%0")
                 (match_operand:SI 2 "general_operand"  " rmi")))]
""
"* return msp430_xorsi_code(insn, operands, NULL);"
  [(set_attr "length" "6")
   (set_attr "cc" "set_n")])


(define_split
 [(set (match_operand:SI 0 "nonimmediate_operand"          "")
       (xor:SI (match_operand:SI 1 "nonimmediate_operand"  "")
	       (match_operand:SI 2 "const_int_operand"     "")))]
 "reload_completed 
  && (halfnibble_integer(operands[2], VOIDmode)
       || halfnibble_constant(operands[2], VOIDmode))
  && INTVAL(operands[2])"
 [(set (match_dup 3) (xor:HI (match_dup 4) (match_dup 5)))]
 "{
      int lo = trunc_int_for_mode(INTVAL(operands[2]),HImode);
      int hi = trunc_int_for_mode(INTVAL(operands[2])>>16,HImode);
      
      if(lo)
      {
	operands[3] = gen_lowpart(HImode, operands[0]);
	operands[4] = gen_lowpart(HImode, operands[1]);
	operands[5] = GEN_INT(lo);
      }
      else if(hi)
      {
	operands[3] = gen_highpart(HImode, operands[0]);
	operands[4] = gen_highpart(HImode, operands[1]);
	operands[5] = GEN_INT(hi);
      }
 }")


(define_peephole2
 [(set (match_operand:HI 0 "nonimmediate_operand_msp430" "") 
       (const_int 0))
  (set (match_dup 0) (xor:HI (match_dup 0) 
			     (const_int -1)))]
 ""
 [(set (match_dup 0) (const_int -1))]
 "")
 

;;============================================================================
;; xor 4 words (64 bits)
;; same as above

(define_expand "xordi3"
  [(set (match_operand:DI 0 "nonimmediate_operand" "")
        (xor:DI (match_operand:DI 1 "nonimmediate_operand" "")   
                (match_operand:DI 2 "general_operand" "")))]
  "" 
  "")

(define_insn "*xordi3_3"
  [(set (match_operand:DI 0 "nonimmediate_operand"          "=rm")
        (xor:DI (match_operand:DI 1 "nonimmediate_operand"  "%0")
                (match_operand:DI 2 "general_operand"  	    " rmi")))]
""
"* return msp430_xordi_code(insn, operands, NULL);"
  [(set_attr "length" "4")
   (set_attr "cc" "set_n")])





;;============================================================================
;;============================================================================
;;============================================================================
;;============================================================================
;;============================================================================
;;============================================================================
;;============================================================================
;;============================================================================
;;============================================================================
;;============================================================================
;; neg
;; same as above

(define_expand "negqi2"
 [(set (match_operand:QI 0 "nonimmediate_operand" "")
   (neg:QI (match_operand:QI 1 "nonimmediate_operand" "")))]
 ""
 "{ emit_insn(gen_one_cmplqi2(operands[0],operands[1]));
 emit_insn(gen_addqi3(operands[0],operands[0],const1_rtx));
 DONE; }")

(define_expand "neghi2"
 [(set (match_operand:HI 0 "nonimmediate_operand" "")
   (neg:HI (match_operand:HI 1 "nonimmediate_operand" "")))]
 ""
 "{ emit_insn(gen_one_cmplhi2(operands[0],operands[1]));
 emit_insn(gen_addhi3(operands[0],operands[0],const1_rtx));
 DONE; }")

(define_expand "negsi2"
 [(set (match_operand:SI 0 "nonimmediate_operand" "")
   (neg:SI (match_operand:SI 1 "nonimmediate_operand" "")))]
 ""

 "{ emit_insn(gen_one_cmplsi2(operands[0],operands[1]));
 emit_insn(gen_addsi3(operands[0],operands[0],const1_rtx));
 DONE; }")

(define_expand "negdi2"
 [(set (match_operand:DI 0 "nonimmediate_operand" "")
   (neg:DI (match_operand:DI 1 "nonimmediate_operand" "")))]
 ""

 "{ emit_insn(gen_one_cmpldi2(operands[0],operands[1]));
 emit_insn(gen_adddi3(operands[0],operands[0],const1_rtx));
 DONE; }")

(define_insn "negsf2"
 [(set (match_operand:SF 0 "nonimmediate_operand" "=r,m")
   (neg:SF (match_operand:SF 1 "nonimmediate_operand" "0,0")))]
 ""
 "xor  #0x8000, %B0"
 [(set_attr "length" "2,3")
 (set_attr "cc" "clobber,clobber")])

;;!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
;;!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
;;!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
;;!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
;;!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
;;!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
;;!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
;;!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
;; not	x = !x
;; ones component

(define_expand "one_cmplqi2"
  [(set (match_operand:QI 0 "nonimmediate_operand_msp430" "")
        (not:QI (match_operand:QI 1 "nonimmediate_operand_msp430" "")))]
  ""
  "")

(define_insn "*one_cmplqi2_2"
  [(set (match_operand:QI 0 "nonimmediate_operand_msp430" 	   "=r,m")
        (not:QI (match_operand:QI 1 "nonimmediate_operand_msp430" " 0, 0")))]
  ""
  "inv.b	%0"
  [(set_attr "length" "1,2")
   (set_attr "cc" "set_czn,set_czn")])


;;============================================================================
;; not HI	x = !x
;; - ones component

(define_expand "one_cmplhi2"
  [(set (match_operand:HI 0 "nonimmediate_operand_msp430" "")
        (not:HI (match_operand:HI 1 "nonimmediate_operand_msp430" "0")))]
  ""
  "")

(define_insn "*one_cmplhi2_2"
  [(set (match_operand:HI 0 "nonimmediate_operand_msp430" 	   "=r, m")
        (not:HI (match_operand:HI 1 "nonimmediate_operand_msp430" " 0, 0")))]
  ""
  "inv	%0"
  [(set_attr "length" "1,2")
   (set_attr "cc" "set_czn,set_czn")])



;;============================================================================
;; not SI	x = !x 
;; - ones component

(define_expand "one_cmplsi2"
  [(set (match_operand:SI 0 "nonimmediate_operand" "")
        (not:SI (match_operand:SI 1 "nonimmediate_operand" "0")))]
  ""
  "")

(define_insn "*one_cmplsi2_2"
  [(set (match_operand:SI 0 "nonimmediate_operand" 	   "=r, m")
        (not:SI (match_operand:SI 1 "nonimmediate_operand" " 0, 0")))]
""
  "inv	%A0
	inv	%B0"
  [(set_attr "length" "2,4")
   (set_attr "cc" "set_n,set_n")])


;;============================================================================
;; not DI	x = !x
;; - ones component

(define_expand "one_cmpldi2"
  [(set (match_operand:DI 0 "nonimmediate_operand" "")
        (not:DI (match_operand:DI 1 "nonimmediate_operand" "0")))]
  ""
  "")

(define_insn "*one_cmpldi2_2"
  [(set (match_operand:DI 0 "nonimmediate_operand" 	   "=r, m")
        (not:DI (match_operand:DI 1 "nonimmediate_operand" " 0, 0")))]
  ""
  "inv	%A0
	inv	%B0
	inv	%C0
	inv	%D0"
  [(set_attr "length" "4,8")
   (set_attr "cc" "set_n,set_n")])



;;============================================================================
;;============================================================================
;;============================================================================
;;============================================================================
;;============================================================================
;;============================================================================
;;============================================================================
;;============================================================================
;;============================================================================
;; abs
;; x = |x|

(define_expand "absqi2"
  [(set (match_operand:QI 0 "nonimmediate_operand" "")
        (abs:QI (match_operand:QI 1 "nonimmediate_operand" "")))]
  ""
  "")

(define_insn "*absqi2_2"
  [(set (match_operand:QI 0 "nonimmediate_operand" 	   "=r,m")
        (abs:QI (match_operand:QI 1 "nonimmediate_operand" " 0, 0")))]
  ""
	"tst.b	%0
	jge	.Leaq%=
	inv.b	%0
	inc.b	%0
.Leaq%=:"
  [(set_attr "length" "4,7")
   (set_attr "cc" "set_czn,set_czn")])


;;============================================================================
;; abs HI	x = |x|
;; 

(define_expand "abshi2"
  [(set (match_operand:HI 0 "nonimmediate_operand" "")
        (abs:HI (match_operand:HI 1 "nonimmediate_operand" "0")))]
  ""
  "")


(define_insn "*abshi2_2"
  [(set (match_operand:HI 0 "nonimmediate_operand" 	   "=r, m")
        (abs:HI (match_operand:HI 1 "nonimmediate_operand" " 0, 0")))]
  ""
	"tst	%0
	jge	.Lae%=
	inv	%0
	inc	%0
.Lae%=:"
  [(set_attr "length" "4,7")
   (set_attr "cc" "set_czn,set_czn")])


;;============================================================================
;; abs SI	x = |x|

(define_expand "abssi2"
  [(set (match_operand:SI 0 "nonimmediate_operand" "")
        (abs:SI (match_operand:SI 1 "nonimmediate_operand" "0")))]
  ""
  "")

(define_insn "*abssi2_2"
  [(set (match_operand:SI 0 "nonimmediate_operand" 	   "=r, m")
        (abs:SI (match_operand:SI 1 "nonimmediate_operand" " 0, 0")))]
  ""
  "* return msp430_emit_abssi(insn, operands,NULL);"
  [(set_attr "length" "7,13")
   (set_attr "cc" "clobber,clobber")])


;;============================================================================
;; abs DI	x = |x|

(define_expand "absdi2"
  [(set (match_operand:DI 0 "nonimmediate_operand" "")
        (abs:DI (match_operand:DI 1 "nonimmediate_operand" "0")))]
  ""
  "")

(define_insn "*absdi2_2"
  [(set (match_operand:DI 0 "nonimmediate_operand" 	   "=r, m")
        (abs:DI (match_operand:DI 1 "nonimmediate_operand" " 0, 0")))]
  ""
  "* return msp430_emit_absdi(insn, operands,NULL);"
  [(set_attr "length" "11,23")
   (set_attr "cc" "clobber,clobber")])

;;============================================================================
;; abs SF

(define_insn "abssf2"
  [(set (match_operand:SF 0 "nonimmediate_operand" "=r,m")
        (abs:SF (match_operand:SF 1 "nonimmediate_operand" "0,0")))]
""
"and	#0x7fff, %B0"
  [(set_attr "length" "2,3")
   (set_attr "cc" "clobber,clobber")])


;; ==========================================================================
;; there are shift helpers

(define_insn "trunchiqi"
  [(set (match_operand:QI 0 "nonimmediate_operand_msp430" "=r,m") 
	(truncate:QI (match_operand:HI 1 "register_operand" "r,r")))]
""
"mov.b	%1, %0"
  [(set_attr "length" "1,2")
   (set_attr "cc" "none,none")])

(define_insn "truncsihi"
  [(set (match_operand:HI 0 "register_operand" "=r") 
	(truncate:HI (match_operand:SI 1 "register_operand" "r")))]
""
"mov	%1, %0"
  [(set_attr "length" "1")
   (set_attr "cc" "none")])


(define_insn "truncsiqi"
  [(set (match_operand:QI 0 "register_operand" "=r") 
	(truncate:QI (match_operand:SI 1 "register_operand" "r")))]
""
"mov.b	%1, %0"
  [(set_attr "length" "1")
   (set_attr "cc" "none")])


(define_insn "truncdiqi"
  [(set (match_operand:QI 0 "register_operand" "=r") 
	(truncate:QI (match_operand:DI 1 "register_operand" "r")))]
""
"mov.b	%1, %0"
  [(set_attr "length" "1")
   (set_attr "cc" "none")])

(define_insn "truncdisi"
  [(set (match_operand:SI 0 "register_operand" "=r") 
	(truncate:SI (match_operand:DI 1 "register_operand" "r")))]
""
"mov	%A1,%A0
	mov    %B1,%B0"
  [(set_attr "length" "2")
   (set_attr "cc" "none")])


(define_expand "rotlhi3"
  [(set (match_operand:HI 0 "nonimmediate_operand"       "")
        (rotate:HI (match_operand:HI 1 "nonimmediate_operand" "")
                   (match_operand:HI 2 "const_int_operand"  "")))]
""
"
  if(INTVAL(operands[2])!=8) FAIL;
")

(define_insn "*rotlhi3"
  [(set (match_operand:HI 0 "nonimmediate_operand"       "=rR,m")
        (rotate:HI (match_operand:HI 1 "nonimmediate_operand" "0,0")
                   (const_int 8)))]
""
"swpb\\t%0"
  [(set_attr "length" "1,2")
     (set_attr "cc" "none")])


;;<< << << << << << << << << << << << << << << << << << << << << << << << <<
;; << << << << << << << << << << << << << << << << << << << << << << << <<
;;<< << << << << << << << << << << << << << << << << << << << << << << << <<
;; << << << << << << << << << << << << << << << << << << << << << << << <<
;;<< << << << << << << << << << << << << << << << << << << << << << << << <<
;; arithmetic shift left

(define_expand "ashlqi3"
  [(set (match_operand:QI 0 "nonimmediate_operand"       "")
        (ashift:QI (match_operand:QI 1 "nonimmediate_operand" "")
                   (match_operand:QI 2 "general_operand"  "")))]
""
"{
   if(!const_int_operand(operands[2],VOIDmode))
   {
     rtx op0,op1;
     
     op0 = force_reg(QImode,operands[0]);
     op1 = force_reg(QImode,operands[1]);
     operands[2] = copy_to_mode_reg(QImode,operands[2]);
     emit_insn(gen_ashlqi3_cnt (op0, op1, operands[2]));
     emit_move_insn(operands[0],op0);
     /*emit_move_insn(operands[1],op1);*/ 
     DONE;
   }
   else if(!register_operand(operands[1], QImode) 
		&& is_shift_better_in_reg(operands))
   {
      operands[1] = copy_to_mode_reg(QImode,operands[1]);
      emit_insn (gen_ashlqi3fnl(operands[0], operands[1], operands[2]));
      DONE;
   }
}")

(define_insn "ashlqi3_cnt"
  [(parallel [(set (match_operand:QI 0 "register_operand"       "=r")
	      (ashift:QI (match_operand:QI 1 "register_operand" "0")
			 (match_operand:QI 2 "register_operand" "")))
	      (clobber (match_dup 2))])]
""
"* return msp430_emit_ashlqi3(insn, operands,NULL);"
  [(set_attr "length" "8")
   (set_attr "cc" "clobber")])

(define_insn "ashlqi3fnl"
  [(set (match_operand:QI 0 "nonimmediate_operand_msp430"       "=rm")
        (ashift:QI (match_operand:QI 1 "nonimmediate_operand_msp430" "0")
                   (match_operand:QI 2 "const_int_operand"  "i")))]
  ""
  "* return msp430_emit_ashlqi3(insn, operands,NULL);"
  [(set_attr "length" "1")
   (set_attr "cc" "clobber")])

;; HImode  ======================================
(define_expand "ashlhi3"
  [(set (match_operand:HI 0 "nonimmediate_operand"       "")
        (ashift:HI (match_operand:HI 1 "nonimmediate_operand" "")
                   (match_operand 2 "general_operand"  "")))]
""
"{ msp430_ashlhi3(operands); DONE; }")

(define_insn "ashlhi3_cnt"
  [(parallel [(set (match_operand:HI 0 "register_operand"       "=r")
	      (ashift:HI (match_operand:HI 1 "register_operand" "0")
			 (match_operand:HI 2 "register_operand" "r")))
	      (clobber (match_dup 2))])]
  ""
  "* return msp430_emit_ashlhi3(insn, operands,NULL);"
  [(set_attr "length" "5")
   (set_attr "cc" "clobber")])

(define_insn "*ashlhi3_1"
  [(set (match_operand:HI 0 "nonimmediate_operand_msp430"       "=r,R,m")
        (ashift:HI (match_operand:HI 1 "nonimmediate_operand_msp430" "0,0,0")
                   (const_int 1)))]
  ""
  "rla\\t%0"
  [(set_attr "length" "1,2,3")
   (set_attr "cc" "clobber")])

(define_insn "*ashlhi3_15"
  [(set (match_operand:HI 0 "nonimmediate_operand_msp430"       "=r,R,m")
        (ashift:HI (match_operand:HI 1 "nonimmediate_operand_msp430" "0,0,0")
                   (const_int 15)))]
  ""
  "rra\\t%0
\\tclr\\t%0
\\trrc\\t%0"
  [(set_attr "length" "3,5,7")
   (set_attr "cc" "clobber")])


;; SImode ======================================

(define_expand "ashlsi3"
  [(set (match_operand:SI 0 "nonimmediate_operand"       "")
        (ashift:SI (match_operand:SI 1 "nonimmediate_operand" "")
                   (match_operand:HI 2 "general_operand"  "")))]
""
"{ msp430_ashlsi3(operands); DONE; }")

(define_insn "ashlsi3_cnt"
  [(parallel [(set (match_operand:SI 0 "register_operand"       "=r")
	      (ashift:SI (match_operand:SI 1 "register_operand" "0")
			 (match_operand:HI 2 "register_operand" "")))
	      (clobber (match_dup 2))])]
""
"* return msp430_emit_ashlsi3(insn, operands,NULL);"
  [(set_attr "length" "8")
   (set_attr "cc" "clobber")])

(define_insn "*ashlsi3_31"
  [(set (match_operand:SI 0 "nonimmediate_operand_msp430"       "=r,R,m")
        (ashift:SI (match_operand:SI 1 "nonimmediate_operand_msp430" "0,0,0")
                   (const_int 31)))]
  ""
"rra\\t%A0
\\tclr\\t%A0
\\tclr\\t%B0
\\trrc\\t%B0"
[(set_attr "length" "4,7,8")
  (set_attr "cc" "clobber")])


(define_insn "*ashlsi3_8"
  [(set (match_operand:SI 0 "nonimmediate_operand_msp430"       "=r,R,m")
        (ashift:SI (match_operand:SI 1 "nonimmediate_operand_msp430" "0,0,0")
                   (const_int 8)))]
  ""
"*{
    if(which_alternative==0)
    {
      return \"xor.b\\t%A0, %B0\\n\\txor\\t%A0, %B0\\n\\tswpb\\t%B0\\n\\tand.b\\t#-1, %A0\\n\\tswpb\\t%A0 \";
    }
    else
    {
      return \"xor.b\\t%A0, %B0\\n\\tclr.b\\t%L0\\n\\txor\\t%A0, %B0\\n\\tswpb\\t%B0\\n\\tclr.b\\t%J0\\n\\tswpb\\t%A0\";
    }
}"
  [(set_attr "length" "5,11,12")
   (set_attr "cc" "clobber")])

(define_insn "*ashlsi3_16"
  [(set (match_operand:SI 0 "nonimmediate_operand_msp430"       "=r,m,r,m")
        (ashift:SI (match_operand:SI 1 "nonimmediate_operand_msp430" "rR,rR,m,m")
                   (const_int 16)))]
""
"mov	%A1, %B0
\tmov	#0, %A0"
[(set_attr "length" "1,2,2,3")
  (set_attr "cc" "clobber")])

(define_insn "*ashlsi3_1"
  [(set (match_operand:SI 0 "nonimmediate_operand_msp430"       "=m,R,r")
        (ashift:SI (match_operand:SI 1 "nonimmediate_operand_msp430" "0,0,0")
                   (const_int 1)))]
  ""
"rla\\t%A0
\\trlc\\t%B0"
[(set_attr "length" "6,5,2")
  (set_attr "cc" "clobber")])

;; DImode ======================================

(define_expand "ashldi3"
  [(set (match_operand:DI 0 "nonimmediate_operand"       "")
        (ashift:DI (match_operand:DI 1 "nonimmediate_operand" "")
                   (match_operand:HI 2 "general_operand"  "")))]
""
"{
   if( !const_int_operand(operands[2],VOIDmode) ||
       INTVAL(operands[2]) > 1)
   {
     rtx op0,op1;
     
     op0 = force_reg(DImode,operands[0]);
     op1 = force_reg(DImode,operands[1]);
     operands[2] = copy_to_mode_reg(HImode,operands[2]);
     emit_insn(gen_ashldi3_cnt (op0, op1, operands[2]));
     emit_move_insn(operands[0],op0);
     /*emit_move_insn(operands[1],op1);*/ 
     DONE;
   }
   else if(!register_operand(operands[1], DImode) 
		&& is_shift_better_in_reg(operands))
   {
      operands[1] = copy_to_mode_reg(DImode,operands[1]);
      emit_insn (gen_ashldi3fnl(operands[0], operands[1], operands[2]));
      DONE;
   }
}")

(define_insn "ashldi3_cnt"
  [(parallel [(set (match_operand:DI 0 "register_operand"       "=r")
	      (ashift:DI (match_operand:DI 1 "register_operand" "0")
			 (match_operand:HI 2 "register_operand" "")))
	      (clobber (match_dup 2))])]
  ""
  "* return msp430_emit_ashldi3(insn, operands,NULL);"
  [(set_attr "length" "8")
   (set_attr "cc" "clobber")])


(define_insn "ashldi3fnl"
  [(set (match_operand:DI 0 "nonimmediate_operand"       "=rm")
        (ashift:DI (match_operand:DI 1 "nonimmediate_operand" "0")
                   (match_operand:HI 2 "const_int_operand"  "i")))]
  ""
  "* return msp430_emit_ashldi3(insn, operands,NULL);"
  [(set_attr "length" "1")
   (set_attr "cc" "clobber")])

;; >> >> >> >> >> >> >> >> >> >> >> >> >> >> >> >> >> >> >> >> >> >> >>
;;  >> >> >> >> >> >> >> >> >> >> >> >> >> >> >> >> >> >> >> >> >> >>      
;; >> >> >> >> >> >> >> >> >> >> >> >> >> >> >> >> >> >> >> >> >> >> >>      
;;  >> >> >> >> >> >> >> >> >> >> >> >> >> >> >> >> >> >> >> >> >> >>      
;; >> >> >> >> >> >> >> >> >> >> >> >> >> >> >> >> >> >> >> >> >> >> >>      
;;  >> >> >> >> >> >> >> >> >> >> >> >> >> >> >> >> >> >> >> >> >> >>     
;; arithmetic shift right

(define_expand "ashrqi3"
  [(set (match_operand:QI 0 "nonimmediate_operand"       "")
        (ashiftrt:QI (match_operand:QI 1 "nonimmediate_operand" "")
                   (match_operand:QI 2 "general_operand"  "")))]
""
"{
   if(!const_int_operand(operands[2],VOIDmode))
   {
     rtx op0,op1;
     
     op0 = force_reg(QImode,operands[0]);
     op1 = force_reg(QImode,operands[1]);
     operands[2] = copy_to_mode_reg(QImode,operands[2]);
     emit_insn(gen_ashrqi3_cnt (op0, op1, operands[2]));
     emit_move_insn(operands[0],op0);
     /*emit_move_insn(operands[1],op1);*/ 
     DONE;
   }
   else if(!register_operand(operands[1], QImode) 
		&& INTVAL(operands[2])>2 
		&& INTVAL(operands[2])!=7)
   {
      operands[1] = copy_to_mode_reg(QImode,operands[1]);
      emit_insn (gen_ashrqi3fnl(operands[0], operands[1], operands[2]));
      DONE;
   }
   else if(INTVAL(operands[2]) == 7)
   {
      /* to do it simple we need a register */
      rtx r1 = gen_reg_rtx(HImode);
      emit_insn(gen_extendqihi2(r1,operands[1]));
      emit_insn(gen_swpb(r1,r1));
      emit_insn(gen_trunchiqi(operands[0],r1));
      DONE;
   }
}")

(define_insn "ashrqi3_cnt"
  [(parallel [(set (match_operand:QI 0 "register_operand"       "=r")
	      (ashiftrt:QI (match_operand:QI 1 "register_operand" "0")
			 (match_operand:QI 2 "register_operand" "")))
	      (clobber (match_dup 2))])]
""
"* return msp430_emit_ashrqi3(insn, operands,NULL);"
  [(set_attr "length" "8")
   (set_attr "cc" "clobber")])

(define_insn "ashrqi3fnl"
  [(set (match_operand:QI 0 "nonimmediate_operand_msp430"       "=rm")
        (ashiftrt:QI (match_operand:QI 1 "nonimmediate_operand_msp430" "0")
                   (match_operand:QI 2 "const_int_operand"  "i")))]
  ""
  "* return msp430_emit_ashrqi3(insn, operands,NULL);"
  [(set_attr "length" "1")
   (set_attr "cc" "clobber")])

;; HImode  ======================================
(define_expand "ashrhi3"
  [(set (match_operand:HI 0 "nonimmediate_operand"       "")
        (ashiftrt:HI (match_operand:HI 1 "nonimmediate_operand" "")
                   (match_operand 2 "general_operand"  "")))]
""
"{msp430_ashrhi3(operands); DONE; }")

(define_insn "ashrhi3_cnt"
  [(parallel [(set (match_operand:HI 0 "register_operand"       "=r")
	      (ashiftrt:HI (match_operand:HI 1 "register_operand" "0")
			 (match_operand:HI 2 "register_operand" "")))
	      (clobber (match_dup 2))])]
""
"* return msp430_emit_ashrhi3(insn, operands,NULL);"
  [(set_attr "length" "5")
   (set_attr "cc" "clobber")])


(define_insn "*ashrhi3_1"
  [(set (match_operand:HI 0 "nonimmediate_operand_msp430"       "=rR,m")
        (ashiftrt:HI (match_operand:HI 1 "nonimmediate_operand_msp430" "0,0")
		     (const_int 1)))]
  ""
  "rra\\t%0"
  [(set_attr "length" "1,2")
   (set_attr "cc" "clobber")])


(define_insn "*ashrhi3_15"
  [(set (match_operand:HI 0 "nonimmediate_operand_msp430"       "=rR,m")
        (ashiftrt:HI (match_operand:HI 1 "nonimmediate_operand_msp430" "0,0")
		     (const_int 15)))]
  ""
  "swpb\\t%0
\\tsxt\\t%0
\\tswpb\\t%0
\\tsxt\\t%0"
  [(set_attr "length" "4,8")
   (set_attr "cc" "clobber")])



;; SImode ======================================

(define_expand "ashrsi3"
  [(set (match_operand:SI 0 "nonimmediate_operand"       "")
        (ashiftrt:SI (match_operand:SI 1 "nonimmediate_operand" "")
                   (match_operand:HI 2 "general_operand"  "")))]
""
"{msp430_ashrsi3(operands);DONE; }")

(define_insn "ashrsi3_cnt"
  [(parallel [(set (match_operand:SI 0 "register_operand"       "=r")
	      (ashiftrt:SI (match_operand:SI 1 "register_operand" "0")
			 (match_operand:HI 2 "register_operand" "")))
	      (clobber (match_dup 2))])]
""
"* return msp430_emit_ashrsi3(insn, operands,NULL);"
  [(set_attr "length" "8")
   (set_attr "cc" "clobber")])

(define_insn "*ashrsi3_1"
  [(set (match_operand:SI 0 "nonimmediate_operand_msp430"       "=rR,m")
        (ashiftrt:SI (match_operand:SI 1 "nonimmediate_operand_msp430" "0,0")
		     (const_int 1)))]
  ""
  "rra\\t%B0
\\trrc\\t%A0"
  [(set_attr "length" "2,4")
   (set_attr "cc" "clobber")])

(define_insn "*ashrsi3_31"
  [(set (match_operand:SI 0 "nonimmediate_operand_msp430"       "=r,R,m")
        (ashiftrt:SI (match_operand:SI 1 "nonimmediate_operand_msp430" "0,0,0")
		     (const_int 31)))]
  ""
"swpb	%B0
\\tsxt	%B0
\\tswpb	%B0
\\tsxt	%B0
\\tmov	%B0, %A0"
  [(set_attr "length" "5,10,10")
   (set_attr "cc" "clobber")])

(define_insn "*ashrsi3_8"
  [(set (match_operand:SI 0 "nonimmediate_operand_msp430"       "=r,R,m")
        (ashiftrt:SI (match_operand:SI 1 "nonimmediate_operand_msp430" "0,0,0")
		     (const_int 8)))]
  ""
"*{
    if(which_alternative==0)
    {
      return \" swpb\\t%A0\\n\\tswpb\\t%B0\\n\\txor.b\\t%B0, %A0\\n\\txor\\t%B0, %A0\\n\\tsxt\\t%B0\";
    }
    else
    {
      return \" swpb\\t%A0\\n\\tswpb\\t%B0\\n\\txor.b\\t%B0, %A0\\n\\tclr.b\\t%J0\\n\\txor\\t%B0, %A0\\n\\tsxt\\t%B0\";
    }
}"
  [(set_attr "length" "5,11,12")
   (set_attr "cc" "clobber")])

;; DImode ======================================

(define_expand "ashrdi3"
  [(set (match_operand:DI 0 "nonimmediate_operand"       "")
        (ashiftrt:DI (match_operand:DI 1 "nonimmediate_operand" "")
                   (match_operand:HI 2 "general_operand"  "")))]
""
"{
   if( !const_int_operand(operands[2],VOIDmode))
   {
     rtx op0,op1;
     
     op0 = force_reg(DImode,operands[0]);
     op1 = force_reg(DImode,operands[1]);
     operands[2] = copy_to_mode_reg(HImode,operands[2]);
     emit_insn(gen_ashrdi3_cnt (op0, op1, operands[2]));
     emit_move_insn(operands[0],op0);
     /*emit_move_insn(operands[1],op1);*/ 
     DONE;
   }
   else if(!register_operand(operands[1], DImode) 
		&& is_shift_better_in_reg(operands))
   {
      operands[1] = copy_to_mode_reg(DImode,operands[1]);
      emit_insn (gen_ashrdi3fnl(operands[0], operands[1], operands[2]));
      DONE;
   }
}")

(define_insn "ashrdi3_cnt"
  [(parallel [(set (match_operand:DI 0 "register_operand"       "=r")
	      (ashiftrt:DI (match_operand:DI 1 "register_operand" "0")
			 (match_operand:HI 2 "register_operand" "")))
	      (clobber (match_dup 2))])]
  ""
  "* return msp430_emit_ashrdi3(insn, operands,NULL);"
  [(set_attr "length" "8")
   (set_attr "cc" "clobber")])


(define_insn "ashrdi3fnl"
  [(set (match_operand:DI 0 "nonimmediate_operand"       "=rm")
        (ashiftrt:DI (match_operand:DI 1 "nonimmediate_operand" "0")
                   (match_operand:HI 2 "const_int_operand"  "i")))]
  ""
  "* return msp430_emit_ashrdi3(insn, operands,NULL);"
  [(set_attr "length" "1")
   (set_attr "cc" "clobber")])





;; >> >> >> >> >> >> >> >> >> >> >> >> >> >> >> >> >> >> >> >> >> >> >> >>
;;  >> >> >> >> >> >> >> >> >> >> >> >> >> >> >> >> >> >> >> >> >> >> >>   
;; >> >> >> >> >> >> >> >> >> >> >> >> >> >> >> >> >> >> >> >> >> >> >> >>   
;;  >> >> >> >> >> >> >> >> >> >> >> >> >> >> >> >> >> >> >> >> >> >> >>   
;; >> >> >> >> >> >> >> >> >> >> >> >> >> >> >> >> >> >> >> >> >> >> >> >>   
;;  >> >> >> >> >> >> >> >> >> >> >> >> >> >> >> >> >> >> >> >> >> >> >>   
;; logical shift right

(define_expand "lshrqi3"
  [(set (match_operand:QI 0 "nonimmediate_operand"       "")
        (lshiftrt:QI (match_operand:QI 1 "nonimmediate_operand" "")
                   (match_operand:QI 2 "general_operand"  "")))]
""
"{
   if(!const_int_operand(operands[2],VOIDmode))
   {
     rtx op0,op1;
     
     op0 = force_reg(QImode,operands[0]);
     op1 = force_reg(QImode,operands[1]);
     operands[2] = copy_to_mode_reg(QImode,operands[2]);
     emit_insn(gen_lshrqi3_cnt (op0, op1, operands[2]));
     emit_move_insn(operands[0],op0);
     /*emit_move_insn(operands[1],op1);*/ 
     DONE;
   }
   else if(!register_operand(operands[1], QImode) 
		&& is_shift_better_in_reg(operands))
   {
      operands[1] = copy_to_mode_reg(QImode,operands[1]);
      emit_insn (gen_lshrqi3fnl(operands[0], operands[1], operands[2]));
      DONE;
   }
}")

(define_expand "lshrqi3_cnt"
  [(parallel [(set (match_operand:QI 0 "register_operand"       "=r")
	      (lshiftrt:QI (match_operand:QI 1 "register_operand" "0")
			 (match_operand:QI 2 "register_operand" "")))
	      (clobber (match_dup 2))])]
""
"")

(define_insn "*lshrqi3_cnt"
  [(parallel [(set (match_operand:QI 0 "register_operand"       "=r")
	      (lshiftrt:QI (match_operand:QI 1 "register_operand" "0")
			 (match_operand:QI 2 "register_operand" "")))
	      (clobber (match_dup 2))])]
""
"* return msp430_emit_lshrqi3(insn, operands,NULL);"
  [(set_attr "length" "8")
   (set_attr "cc" "clobber")])

(define_insn "lshrqi3fnl"
  [(set (match_operand:QI 0 "nonimmediate_operand_msp430"       "=rm")
        (lshiftrt:QI (match_operand:QI 1 "nonimmediate_operand_msp430" "0")
                   (match_operand:QI 2 "const_int_operand"  "i")))]
  ""
  "* return msp430_emit_lshrqi3(insn, operands,NULL);"
  [(set_attr "length" "1")
   (set_attr "cc" "clobber")])

;; HImode  ======================================

(define_insn "clrc"
 [(unspec:HI [(const_int 123454321)] 30)]
""
 "clrc"
[(set_attr "length" "1")
  (set_attr "cc" "clobber")])

 
(define_expand "lshrhi3"
  [(set (match_operand:HI 0 "nonimmediate_operand"       "")
        (lshiftrt:HI (match_operand:HI 1 "nonimmediate_operand" "")
                   (match_operand 2 "general_operand"  "")))]
""
"{msp430_lshrhi3(operands); DONE; }")

(define_insn "lshrhi3_cnt"
  [(parallel [(set (match_operand:HI 0 "register_operand"       "=r")
	      (lshiftrt:HI (match_operand:HI 1 "register_operand" "0")
			 (match_operand:HI 2 "register_operand" "")))
	      (clobber (match_dup 2))])]
""
"* return msp430_emit_lshrhi3(insn, operands,NULL);"
  [(set_attr "length" "5")
   (set_attr "cc" "clobber")])

(define_insn "*lshrhi3_15"
  [(set (match_operand:HI 0 "nonimmediate_operand_msp430"       "=r,R,m")
        (lshiftrt:HI (match_operand:HI 1 "nonimmediate_operand_msp430" "0,0,0")
		     (const_int 15)))]
  ""
  "rla\\t%0
\\tclr\\t%0
\\trlc\\t%0"
  [(set_attr "length" "3,6,8")
   (set_attr "cc" "clobber")])


(define_insn "*lshrhi3_1"
  [(set (match_operand:HI 0 "nonimmediate_operand_msp430"       "=rR,m")
        (lshiftrt:HI (match_operand:HI 1 "nonimmediate_operand_msp430" "0,0")
		     (const_int 1)))]
  ""
"clrc
\\trrc\\t%0"
  [(set_attr "length" "2,3")
   (set_attr "cc" "clobber")])

;; SImode ======================================

(define_expand "lshrsi3"
  [(set (match_operand:SI 0 "nonimmediate_operand"       "")
        (lshiftrt:SI (match_operand:SI 1 "nonimmediate_operand" "")
                   (match_operand:HI 2 "general_operand"  "")))]
""
"{ msp430_lshrsi3(operands); DONE; }")

(define_insn "lshrsi3_cnt"
  [(parallel [(set (match_operand:SI 0 "register_operand"       "=r")
	      (lshiftrt:SI (match_operand:SI 1 "register_operand" "0")
			 (match_operand:HI 2 "register_operand" "")))
	      (clobber (match_dup 2))])]
""
"* return msp430_emit_lshrsi3(insn, operands,NULL);"
  [(set_attr "length" "8")
   (set_attr "cc" "clobber")])


(define_insn "*lshrsi3_31"
  [(set (match_operand:SI 0 "nonimmediate_operand_msp430"       "=r,R,m")
        (lshiftrt:SI (match_operand:SI 1 "nonimmediate_operand_msp430" "0,0,0")
		     (const_int 31)))]
  ""
"rla	%B0
\\tclr	%B0
\\tclr	%A0
\\trlc	%A0"
  [(set_attr "length" "4,9,10")
   (set_attr "cc" "clobber")])


(define_insn "*lshrsi3_8"
  [(set (match_operand:SI 0 "nonimmediate_operand_msp430"       "=r,R,m")
        (lshiftrt:SI (match_operand:SI 1 "nonimmediate_operand_msp430" "0,0,0")
		     (const_int 8)))]
  ""
"*{
    if(which_alternative==0)
    {
      return \"swpb\\t%A0\\n\\tswpb\\t%B0\\n\\txor.b\\t%B0, %A0\\n\\txor\\t%B0, %A0\\n\\tand.b\\t#-1, %B0\";
    }
    else
    {
      return \"swpb\\t%A0\\n\\tswpb\\t%B0\\n\\txor.b\\t%B0, %A0\\n\\tclr.b\\t%J0\\n\\txor\\t%B0, %A0\\n\\tclr.b\\t%L0\";
    }
}"
  [(set_attr "length" "5,11,12")
   (set_attr "cc" "clobber")])

(define_insn "*lshrsi3_1"
  [(set (match_operand:SI 0 "nonimmediate_operand_msp430"       "=r,R,m")
        (lshiftrt:SI (match_operand:SI 1 "nonimmediate_operand_msp430" "0,0,0")
		     (const_int 1)))]
  ""
"clrc
\\trrc\\t%B0
\\trrc\\t%A0"
  [(set_attr "length" "3,4,5")
   (set_attr "cc" "clobber")])

;; DImode ======================================

(define_expand "lshrdi3"
  [(set (match_operand:DI 0 "nonimmediate_operand"       "")
        (lshiftrt:DI (match_operand:DI 1 "nonimmediate_operand" "")
                   (match_operand:HI 2 "general_operand"  "")))]
""
"{
   if( !const_int_operand(operands[2],VOIDmode))
   {
     rtx op0,op1;
     
     op0 = force_reg(DImode,operands[0]);
     op1 = force_reg(DImode,operands[1]);
     operands[2] = copy_to_mode_reg(HImode,operands[2]);
     emit_insn(gen_lshrdi3_cnt (op0, op1, operands[2]));
     emit_move_insn(operands[0],op0);
     /*emit_move_insn(operands[1],op1);*/ 
     DONE;
   }
   else if(!register_operand(operands[1], DImode) 
		&& is_shift_better_in_reg(operands))
   {
      operands[1] = copy_to_mode_reg(DImode,operands[1]);
      emit_insn (gen_lshrdi3fnl(operands[0], operands[1], operands[2]));
      DONE;
   }
}")

(define_insn "lshrdi3_cnt"
  [(parallel [(set (match_operand:DI 0 "register_operand"       "=r")
	      (lshiftrt:DI (match_operand:DI 1 "register_operand" "0")
			 (match_operand:HI 2 "register_operand" "")))
	      (clobber (match_dup 2))])]
  ""
  "* return msp430_emit_lshrdi3(insn, operands,NULL);"
  [(set_attr "length" "8")
   (set_attr "cc" "clobber")])


(define_insn "lshrdi3fnl"
  [(set (match_operand:DI 0 "nonimmediate_operand"       "=rm")
        (lshiftrt:DI (match_operand:DI 1 "nonimmediate_operand" "0")
                   (match_operand:HI 2 "const_int_operand"  "i")))]
  ""
  "* return msp430_emit_lshrdi3(insn, operands,NULL);"
  [(set_attr "length" "1")
   (set_attr "cc" "clobber")])






;; xx<---x xx<---x xx<---x xx<---x xx<---x xx<---x xx<---x xx<---x xx<---x
;; xx<---x xx<---x xx<---x xx<---x xx<---x xx<---x xx<---x xx<---x xx<---x
;; xx<---x xx<---x xx<---x xx<---x xx<---x xx<---x xx<---x xx<---x xx<---x
;; xx<---x xx<---x xx<---x xx<---x xx<---x xx<---x xx<---x xx<---x xx<---x
;; xx<---x xx<---x xx<---x xx<---x xx<---x xx<---x xx<---x xx<---x xx<---x
;; xx<---x xx<---x xx<---x xx<---x xx<---x xx<---x xx<---x xx<---x xx<---x
;; sign extend

(define_insn "extendqihi2"
  [(set (match_operand:HI 0 "nonimmediate_operand" "=rm,rm")
        (sign_extend:HI (match_operand:QI 1 "general_operand" "0,*rmi")))]
  ""
  "* return signextendqihi(insn, operands,NULL);"
  [(set_attr "length" "2,2")
   (set_attr "cc" "set_n,set_n")])

(define_insn "extendqisi2"
  [(set (match_operand:SI 0 "nonimmediate_operand" "=rm,rm")
        (sign_extend:SI (match_operand:QI 1 "general_operand" "0,*rmi")))]
  ""
  "* return signextendqisi(insn, operands,NULL);"
  [(set_attr "length" "6,6")
   (set_attr "cc" "set_n,set_n")])

(define_insn "extendqidi2"
  [(set (match_operand:DI 0 "nonimmediate_operand" "=rm,rm")
        (sign_extend:DI (match_operand:QI 1 "general_operand" "0,*rmi")))]
  ""
  "* return signextendqidi(insn, operands,NULL);"
  [(set_attr "length" "6,6")
   (set_attr "cc" "set_n,set_n")])

(define_insn "extendhisi2"
  [(set (match_operand:SI 0 "nonimmediate_operand" "=rm,rm")
        (sign_extend:SI (match_operand:HI 1 "general_operand" "0,*rmi")))] 
  ""
  "* return signextendhisi(insn, operands,NULL);"
  [(set_attr "length" "6,6")  
   (set_attr "cc" "set_n,set_n")])      

(define_insn "extendhidi2"
  [(set (match_operand:DI 0 "nonimmediate_operand" "=rm,rm")
        (sign_extend:DI (match_operand:HI 1 "general_operand" "0,*rmi")))]
  ""
  "* return signextendhidi(insn, operands,NULL);"
  [(set_attr "length" "6,6")  
   (set_attr "cc" "set_n,set_n")])

(define_insn "extendsidi2"
  [(set (match_operand:DI 0 "nonimmediate_operand" "=rm,rm")
        (sign_extend:DI (match_operand:SI 1 "general_operand" "0,*rmi")))]
  ""
  "* return signextendsidi(insn, operands,NULL);"
  [(set_attr "length" "6,6")
   (set_attr "cc" "set_n,set_n")])

;; xx<---0 xx<---0 xx<---0 xx<---0 xx<---0 xx<---0 xx<---0 xx<---0 xx<---0
;; xx<---0 xx<---0 xx<---0 xx<---0 xx<---0 xx<---0 xx<---0 xx<---0 xx<---0
;; xx<---0 xx<---0 xx<---0 xx<---0 xx<---0 xx<---0 xx<---0 xx<---0 xx<---0
;; xx<---0 xx<---0 xx<---0 xx<---0 xx<---0 xx<---0 xx<---0 xx<---0 xx<---0
;; xx<---0 xx<---0 xx<---0 xx<---0 xx<---0 xx<---0 xx<---0 xx<---0 xx<---0
;; xx<---0 xx<---0 xx<---0 xx<---0 xx<---0 xx<---0 xx<---0 xx<---0 xx<---0
;; zero extend

(define_insn "zero_extendqihi2"
  [(set (match_operand:HI 0 "nonimmediate_operand" "=rm,rm")
        (zero_extend:HI (match_operand:QI 1 "general_operand" "0,*rmi")))]
  ""
  "* return zeroextendqihi(insn, operands,NULL);"
  [(set_attr "length" "2,2")
   (set_attr "cc" "clobber,clobber")])

(define_insn "zero_extendqisi2"
  [(set (match_operand:SI 0 "nonimmediate_operand" "=rm,rm")
        (zero_extend:SI (match_operand:QI 1 "general_operand" "0,*rmi")))]
  ""
  "* return zeroextendqisi(insn, operands,NULL);"
  [(set_attr "length" "6,6")
   (set_attr "cc" "clobber,clobber")])

(define_insn "zero_extendqidi2"
  [(set (match_operand:DI 0 "nonimmediate_operand" "=rm,rm")
        (zero_extend:DI (match_operand:QI 1 "general_operand" "0,*rmi")))]
  ""
  "* return zeroextendqidi(insn, operands,NULL);"
  [(set_attr "length" "6,6")
   (set_attr "cc" "clobber,clobber")])


(define_insn "zero_extendhisi2"
  [(set (match_operand:SI 0 "nonimmediate_operand" "=rm,rm")
        (zero_extend:SI (match_operand:HI 1 "general_operand" "0,*rmi")))] 
  ""
  "* return zeroextendhisi(insn, operands,NULL);"
  [(set_attr "length" "6,6")  
   (set_attr "cc" "clobber,clobber")])      

(define_insn "zero_extendhidi2"
  [(set (match_operand:DI 0 "nonimmediate_operand" "=rm,rm")
        (zero_extend:DI (match_operand:HI 1 "general_operand" "0,*rmi")))]
  ""
  "* return zeroextendhidi(insn, operands,NULL);"
  [(set_attr "length" "6,6")  
   (set_attr "cc" "clobber,clobber")])

(define_insn "zero_extendsidi2"
  [(set (match_operand:DI 0 "nonimmediate_operand" "=rm,rm")
        (zero_extend:DI (match_operand:SI 1 "general_operand" "0,*rmi")))]
  ""
  "* return zeroextendsidi(insn, operands,NULL);"
  [(set_attr "length" "6,6")  
   (set_attr "cc" "clobber,clobber")])

;; =====================================================================
;; single bit extract
;; as soon as all operatoins performed on io registers
;; let use only QImode. 

(define_expand "extv"
 [(set (match_operand:QI 0 "nonimmediate_operand_msp430" "")
       (sign_extract:QI (match_operand:QI 1 "nonimmediate_operand_msp430" "") 
			(match_operand 2 "const_int_operand" "")
			(match_operand 3 "const_int_operand" "")))]
""
"{
    if(INTVAL(operands[2]) != 1 || INTVAL(operands[3]) <= 0)
      FAIL;
}")

(define_insn "*extv" 
 [(set (match_operand:QI 0 "nonimmediate_operand_msp430" "=r,r,r,r,m,m,m,m")
       (sign_extract:QI (match_operand:QI 1 "nonimmediate_operand_msp430" "r,m,r,m,r,m,r,m") 
			(const_int 1)
			(match_operand 2 "const_int_operand" "P,P,i,i,P,P,i,i")))]
 ""
"* {
  operands[2] = GEN_INT(1<<INTVAL(operands[2]));
  return 	\"bit.b\\t%2, %1\\n\"
		\"\\tclr.b\\t%0\\n\"
		\"\\tadc.b\\t%0\";
}"
  [(set_attr "length" "3,4,4,5,5,6,6,7")
     (set_attr "cc" "clobber")])

(define_expand "extzv"
 [(set (match_operand:QI 0 "nonimmediate_operand_msp430" "")
       (zero_extract:QI (match_operand:QI 1 "nonimmediate_operand_msp430" "") 
			(match_operand 2 "const_int_operand" "")
			(match_operand 3 "const_int_operand" "")))]
""
"{
    if(INTVAL(operands[2]) != 1 || INTVAL(operands[3]) <= 0) 
      FAIL;
}")

(define_insn "*extzv" 
 [(set (match_operand:QI 0 "nonimmediate_operand_msp430" "=r,r,r,r,m,m,m,m")
       (zero_extract:QI (match_operand:QI 1 "nonimmediate_operand_msp430" "r,m,r,m,r,m,r,m")
			(match_operand 2 "const_int_operand" "")
			(match_operand 3 "const_int_operand" "P,P,i,i,P,P,i,i")))]
 "INTVAL(operands[2]) == 1"
"* {
  operands[3] = GEN_INT(1<<INTVAL(operands[3]));
  return 	\"bit.b\\t%3, %1\\n\"
		\"\\tclr.b\\t%0\\n\"
		\"\\tadc.b\\t%0\";
}"
  [(set_attr "length" "3,4,4,5,5,6,6,7")
     (set_attr "cc" "clobber")])



;;=======================================================================
;;  various BRANCH insns...
;;
;;

;; Unconditional jump instruction.
(define_insn "jump"
  [(set (pc) (label_ref (match_operand 0 "" "")))]
  ""
  "*
{
  int dist = msp430_jump_dist(operands[0],insn);
  if (dist<500 && dist>-500)
    return \"jmp	%0\";
  return \"br	#%0\"; 
}"
  [(set_attr "length" "2")
   (set_attr "cc" "none")])


(define_insn "explicit_br"
  [(unspec_volatile:HI [(const_int 0)] UNSPEC_EXPLICIT_BR)
   (match_operand:HI 0 "immediate_operand" "")]
   ""
   "br	%0"
  [(set_attr "length" "2")
   (set_attr "cc" "none")]
)

   
; indirect jump
(define_expand "indirect_jump"
  [(set (pc) (match_operand:HI 0 "nonimmediate_operand" ""))]
  ""
  "")

(define_insn "*indirect_jump_idx"
  [(set (pc) (match_operand:HI 0 "memory_operand" "m"))]
  "indexed_location(operands[0])"
  "br	@%E0"
  [(set_attr "length" "1")
   (set_attr "cc" "none")])

(define_insn "*indirect_jump_mem"
  [(set (pc) (match_operand:HI 0 "memory_operand" "m"))]
  "!indexed_location(operands[0])"
  "br	%0"
  [(set_attr "length" "2")
   (set_attr "cc" "none")])


(define_insn "*indirect_jump_reg"
  [(set (pc) (match_operand:HI 0 "register_operand" "r"))]
  ""
  "br	%0"
  [(set_attr "length" "1")
   (set_attr "cc" "none")])


;;=======================================================================
;;=======================================================================
;;=======================================================================
;;=======================================================================


;;=======================================================================
;;
;;    CASE
;;

/*
(define_expand "casesi"
  [(set (match_dup 6)  
        (minus:HI (subreg:HI (match_operand:SI 0 "register_operand" "") 0)
                  (match_operand:HI 1 "register_operand" "")))
   (parallel [(set (cc0)
                   (compare (match_dup 6)
                            (match_operand:HI 2 "register_operand" "")))])
   (set (pc)
        (if_then_else (gtu (cc0)
                           (const_int 0))
                      (label_ref (match_operand 4 "" ""))
                      (pc)))
   (set (match_dup 6)
        (plus:HI (match_dup 6) (label_ref (match_operand:HI 3 "" ""))))

   (parallel [(set (pc) (unspec:HI [(match_dup 6)] 1))
              (use (label_ref (match_dup 3)))
              (clobber (match_dup 6))])]
  ""
  " 
{   
  operands[6] = gen_reg_rtx (HImode);
}")

*/ 

;; Table helper
(define_insn "tablejump"
  [(set (pc) (match_operand:HI 0 "general_operand" "rRP,i,m"))
   (use (label_ref (match_operand 1 "" "")))]
  ""
  "br	%0	;	%1"
 [(set_attr "length" "1,2,2")   
   (set_attr "cc" "clobber")])


;; =============================================================
;; match De Morgan's law
(define_insn "nandqi"
  [(set (match_operand:QI 0 "nonimmediate_operand_msp430" "=r,m,m,r")
             (and:QI (not:QI (match_operand:QI 1 "general_operand_msp430" "rRP,mi,rRP,mi"))
	                     (match_operand:QI 2 "nonimmediate_operand_msp430" "0,0,0,0")))]
""
"bic.b       %1, %0"
[(set_attr "length" "1,3,2,2")
  (set_attr "cc" "none")])

(define_insn "nandhi"
  [(set (match_operand:HI 0 "nonimmediate_operand_msp430" "=r,m,m,r")
             (and:HI (not:HI (match_operand:HI 1 "general_operand_msp430" "rRP,mi,rRP,mi"))
	                     (match_operand:HI 2 "nonimmediate_operand_msp430" "0,0,0,0")))]
""
"bic %1, %0"
[(set_attr "length" "1,3,2,2")
  (set_attr "cc" "none")])

(define_insn "nandsi"
  [(set (match_operand:SI 0 "nonimmediate_operand_msp430" "=r,m,m,r,r,m")
            (and:SI (not:SI (match_operand:SI 1 "general_operand_msp430" "rP,mi,rP,mi,R,R"))
	                    (match_operand:SI 2 "nonimmediate_operand_msp430" "0,0,0,0,0,0")))]
""
"bic    %A1, %A0
\\tbic  %B1, %B0"
[(set_attr "length" "2,6,4,4,3,5")
  (set_attr "cc" "none")])



;; =============================================================
;; PEEPHOLES

;; a &= ~b;

(define_insn "*bit_clear"
 [(set (match_operand:HI 0 "nonimmediate_operand_msp430" "=r,m,m,r") 
       (unspec_volatile:HI [(match_operand:HI 1 "general_operand_msp430" "rRP,mi,rRP,mi")] 40))]
""
"bic	%1, %0"
  [(set_attr "length" "1,3,2,2")
     (set_attr "cc" "none")])

(define_insn "bic_sr_irq"
  [(set (match_operand:HI 0 "nonimmediate_operand_msp430" "=r,m,m,r")
           (unspec_volatile:HI [(match_operand:HI 1 "general_operand_msp430" "rRP,mi,rRP,mi")
				(const_int 4100001)] 41))]
  ""
  "bic    %1, %0"
    [(set_attr "length" "1,3,2,2")
         (set_attr "cc" "none")])

(define_insn "bis_sr_irq"
  [(set (match_operand:HI 0 "nonimmediate_operand_msp430" "=r,m,m,r")
           (unspec_volatile:HI [(match_operand:HI 1 "general_operand_msp430" "rRP,mi,rRP,mi")
				 (const_int 4200002)] 42))]
  ""
  "bis    %1, %0"
    [(set_attr "length" "1,3,2,2")
         (set_attr "cc" "none")])

(define_insn "get_frame_address"
  [(set (match_operand:HI 0 "nonimmediate_operand_msp430" "=r,m,m,r")
           (unspec_volatile:HI [(match_operand:HI 1 "general_operand_msp430" "rRP,mi,rRP,mi")
				 (const_int 4300003)] 43))]
  ""
  "mov	%1, %0"
    [(set_attr "length" "1,3,2,2")
         (set_attr "cc" "none")])

;; these two for:
;; (ulong) x = (ulong) func() << 16;
;; x |= func();
;; func() is uint
;;

;; do not check for zeros here, cause this insn already issued.
(define_peephole2
   [(set (match_operand:SI 1 "register_operand" "")
         (sign_extend:SI (match_operand:HI 0 "register_operand" "")))
    (set (match_dup 1) (ashift:SI (match_dup 1) (const_int 16)))]
""
  [(set (subreg:HI (match_dup 1) 2) (match_dup 0))]
"")

(define_peephole2
   [(set (match_operand:SI 1 "register_operand" "")
         (zero_extend:SI (match_operand:HI 0 "register_operand" "")))
    (set (match_dup 1) (ashift:SI (match_dup 1) (const_int 16)))]
""
  [(set (subreg:HI (match_dup 1) 2) (match_dup 0))]
"")

(define_peephole2
   [(set (match_operand:SI 0 "register_operand" "")
         (zero_extend:SI (match_operand:HI 1 "register_operand" "")))
    (set (match_operand:SI 2 "register_operand" "")
         (ior:SI (match_dup 2) (match_dup 0)))]
"dead_or_set_in_peep(1,insn, operands[0])"
  [(set (subreg:HI (match_dup 2) 0) 
        (ior:HI (subreg:HI (match_dup 2) 0) (match_dup 1)))]
"")

(define_peephole2
   [(set (match_operand:HI 0 "register_operand" "")  
         (match_operand:HI 1 "general_operand_msp430" ""))
    (set (match_operand:SI 2 "register_operand" "")
         (zero_extend:SI (match_dup 0)))
    (set (match_operand:SI 3 "register_operand" "") 
         (ior:SI (match_dup 3) (match_dup 2)))]
"dead_or_set_in_peep(2,insn, operands[0])"
  [(set (subreg:HI (match_dup 3) 0) 
        (ior:HI (subreg:HI (match_dup 3) 0) (match_dup 1)))]
"")


;; (ulong) x = (ulong) f >> 16;
;;
(define_peephole2
  [(set (match_operand:SI 0 "register_operand" "") 
	(match_operand:SI 1 "register_operand" "")) 
   (set (match_dup 0) 
	(lshiftrt:SI (match_dup 0) 
		     (const_int 16)))]
""
[(set (subreg:HI (match_dup 0) 0) (subreg:HI (match_dup 1) 2))
 (set (subreg:HI (match_dup 0) 2) (const_int 0))]
"")


(define_peephole2
  [(set (match_operand:SI 0 "register_operand" "") 
	(match_operand:SI 1 "register_operand" ""))
   (set (match_operand:SI 2 "register_operand" "")
	(ior:SI (match_dup 2) (match_dup 0)))]
"dead_or_set_in_peep(1,insn, operands[0])"
  [(set (match_dup 2) (ior:SI (match_dup 2) 
		              (match_dup 1)))]
"")


(define_peephole2
  [(set (match_operand:SI 0 "register_operand" "") 
	(zero_extend:SI (match_operand:HI 1 "general_operand" "")))
   (set (match_dup 0) 
	(ashift:SI (match_dup 0) (const_int 16)))]
""
  [(set (subreg:HI (match_dup 0) 2) (match_dup 1))]
"")


;; shift right & set
(define_peephole2
  [(set (match_operand:HI 0 "register_operand" "") 
	(match_operand:HI 1 "register_operand" ""))
   (set (match_operand:HI 2 "register_operand" "")
	(match_dup 0))
   (set (match_dup 2) 
	(and:HI (match_dup 2) 
		(match_operand 3 "const_int_operand" "")))
   (set (match_dup 2) 
	(lshiftrt:HI (match_dup 2) 
		     (match_operand 4 "const_int_operand" ""))) 
   (set (match_dup 1) (match_dup 2))]
"dead_or_set_in_peep(4,insn, operands[2])"
  [(set (match_dup 0) (match_dup 1))
   (set (match_dup 1)  
	(and:HI (match_dup 1)  
		(match_dup 3))) 
   (set (match_dup 1) 
	(lshiftrt:HI (match_dup 1) 
		     (match_dup 4)))] 
"")

;; shift left and set
(define_peephole2
  [(set (match_operand:HI 0 "register_operand" "") 
	(match_operand:HI 1 "register_operand" ""))
   (set (match_operand:HI 2 "register_operand" "")
	(match_dup 0))
   (set (match_dup 2) 
	(and:HI (match_dup 2) 
		(match_operand 3 "const_int_operand" "")))
   (set (match_dup 2) 
	(ashift:HI (match_dup 2) 
		   (match_operand 4 "const_int_operand" ""))) 
   (set (match_dup 1) (match_dup 2))]
"dead_or_set_in_peep(4,insn, operands[2])"
  [(set (match_dup 0) (match_dup 1))
   (set (match_dup 1)  
	(and:HI (match_dup 1)  
		(match_dup 3))) 
   (set (match_dup 1) 
	(ashift:HI (match_dup 1) 
		   (match_dup 4)))] 
"")


;;
;; these for some shifts and stuff.
;; every peephole saves up to 4 bytes.
;;

(define_insn "*addc_reg"
  [(set (match_operand:HI 0 "register_operand" "=r,r") 
	(unspec:HI [(match_operand:HI 1 "register_operand" "%0,0") 
		    (match_operand:HI 2 "general_operand_msp430" "rP,mi")] 0))]
""
"addc	%2, %0"
[(set_attr "length" "1,2")
   (set_attr "cc" "clobber,clobber")])


(define_insn "*addc_any"
  [(set (match_operand:HI 0 "nonimmediate_operand_msp430" "=m,m") 
	(unspec:HI [(match_operand:HI 1 "nonimmediate_operand_msp430" "%0,0") 
		    (match_operand:HI 2 "general_operand_msp430" "rP,mi")] 5))]
""
"addc	%2, %0"
[(set_attr "length" "2,3")
   (set_attr "cc" "clobber,clobber")])

(define_peephole2
  [(set (match_operand:HI 0 "register_operand" "") 
	(match_operand:HI 1 "general_operand_msp430" ""))
   (set (match_operand:SI 2 "register_operand" "") 
	(zero_extend:SI (match_dup 0)))
   (set (match_operand:SI 3 "register_operand" "") 
	(plus:SI (match_dup 3) (match_dup 2)))]
"dead_or_set_in_peep(2,insn, operands[2])"
  [(set (subreg:HI (match_dup 3) 0) 
	(plus:HI (subreg:HI (match_dup 3) 0) 
		 (match_dup 1)))
   (set (subreg:HI (match_dup 3) 2) 
	(unspec:HI [(subreg:HI (match_dup 3) 2) (const_int 0)] 0))]
"")

(define_peephole2
  [(set (match_operand:HI 0 "register_operand" "") 
	(match_operand:HI 1 "general_operand_msp430" ""))
   (set (match_operand:SI 2 "register_operand" "") 
	(zero_extend:SI (match_dup 0)))
   (set (match_operand:SI 3 "nonimmediate_operand_msp430" "") 
	(plus:SI (match_dup 3) (match_dup 2)))]
"dead_or_set_in_peep(2,insn, operands[2])"
  [(set (subreg:HI (match_dup 3) 0) 
	(plus:HI (subreg:HI (match_dup 3) 0) 
		 (match_dup 1)))
   (set (subreg:HI (match_dup 3) 2) 
	(unspec:HI [(subreg:HI (match_dup 3) 2) (const_int 0)] 5))]
"")

(define_peephole2
  [(set (match_operand:HI 0 "register_operand" "")
        (match_operand:HI 1 "general_operand_msp430" ""))
   (set (match_operand:SI 2 "nonimmediate_operand_msp430" "")
        (zero_extend:SI (match_dup 0)))]
"dead_or_set_in_peep(1,insn, operands[0])"
  [(set (subreg:HI (match_dup 2) 0) (match_dup 1))
   (set (subreg:HI (match_dup 2) 2) (const_int 0))]
"")

;;
;; these are for redudant moves.
;;

(define_peephole2
  [(set (match_operand:SI 0 "register_operand" "") 
	(match_operand:SI 1 "nonimmediate_operand_msp430" ""))
   (set (match_operand:SI 2 "register_operand" "") 
	(ior:SI (match_dup 2) (match_dup 0)))
   (set (match_dup 1) (match_dup 2))]
"dead_or_set_in_peep(1,insn, operands[0])"
  [(set (match_dup 1) (ior:SI (match_dup 1) (match_dup 2)))]
"")

(define_peephole2
  [(set (match_operand:SI 0 "register_operand" "")
	(ior:SI (match_dup 0) 
		(match_operand:SI 1 "register_operand" "")))
   (set (match_dup 1) (match_dup 0))]
"dead_or_set_in_peep(1,insn, operands[0])"
  [(set (match_dup 1) (ior:SI (match_dup 1) (match_dup 0)))]
"")

(define_peephole2
  [(set (match_operand:HI 0 "register_operand" "") 
	(match_operand:HI 1 "register_operand" ""))
   (set (match_dup 0) 
	(not:HI (match_dup 0)))
   (set (match_operand:HI 2 "register_operand" "") 
	(match_dup 0))]
"dead_or_set_in_peep(2,insn, operands[0]) && dead_or_set_in_peep(0,insn, operands[1])"
  [(set (match_dup 1) (not:HI (match_dup 1)))
   (set (match_dup 2) (match_dup 1))]
"")

(define_peephole2
  [(set (match_operand:SI 0 "register_operand" "") 
	(match_operand:SI 1 "register_operand" ""))
   (set (match_dup 0) 
	(not:SI (match_dup 0)))
   (set (match_operand:SI 2 "register_operand" "") 
	(match_dup 0))]
"dead_or_set_in_peep(2,insn, operands[0]) && dead_or_set_in_peep(0,insn, operands[1])"
  [(set (match_dup 1) (not:SI (match_dup 1)))
   (set (match_dup 2) (match_dup 1))]
"")

(define_peephole2
  [(set (match_operand:SF 0 "register_operand" "")
	(match_operand:SF 1 "general_operand_msp430" ""))
   (set (match_operand:SF 2 "nonimmediate_operand_msp430" "") 
	(match_dup 0))]
"dead_or_set_in_peep(1,insn, operands[0])"
  [(set (match_dup 2) (match_dup 1))]
"")


(define_peephole2
  [(set (match_operand:SI 0 "register_operand" "")
	(match_operand:SI 1 "general_operand_msp430" ""))
   (set (match_operand:SI 2 "nonimmediate_operand_msp430" "") 
	(match_dup 0))]
"dead_or_set_in_peep(1,insn, operands[0])"
  [(set (match_dup 2) (match_dup 1))]
"")

(define_peephole2
  [(set (match_operand:HI 0 "register_operand" "")
        (match_operand:HI 1 "general_operand_msp430" ""))
   (set (match_operand:HI 2 "nonimmediate_operand_msp430" "")
        (match_dup 0))]
"dead_or_set_in_peep(1,insn, operands[0])"
  [(set (match_dup 2) (match_dup 1))]
"")



;; =========================================================================
;;  This one for bit tests like:
;;	volatile long a;
;;	while(a&CONST_HALFNIBBLE) ;

(define_insn "*bittest_lo"
  [(set (cc0) 
	(unspec:SI [(match_operand:SI 0 "nonimmediate_operand_msp430" "r,r,m,m") 
		    (match_operand:SI 1 "general_operand_msp430" "rPR,mi,rPR,mi")] 1))]
""
"bit	%A1,%A0"
[(set_attr "length" "1,2,2,3")
   (set_attr "cc" "compare,compare,compare,compare")])

(define_insn "*bittest_hi"
  [(set (cc0) 
	(unspec:SI [(match_operand:SI 0 "nonimmediate_operand_msp430" "r,r,m,m") 
		    (match_operand:SI 1 "general_operand_msp430" "rPR,mi,rPR,mi")] 2))]
""
"bit	%B1,%B0"
[(set_attr "length" "1,2,2,3")
   (set_attr "cc" "compare,compare,compare,compare")])

(define_peephole2
 [(set (match_operand:SI 0 "register_operand" "")
       (match_operand:SI 1 "nonimmediate_operand_msp430" ""))
  (set (match_dup 0) (and:SI (match_dup 0) 
                             (match_operand 2 "const_int_operand" "")))
  (set (pc) 
       (if_then_else (match_operator:SI 3 "equality_operator"
		      [(match_dup 0) (const_int 0)]) 
       (label_ref (match_operand 4 "" ""))
       (pc)))]
"(halfnibble_integer(operands[2], VOIDmode) 
   || halfnibble_constant(operands[2], VOIDmode))
        && dead_or_set_in_peep(2,insn, operands[0])
	&& which_nibble(INTVAL(operands[2])) == 0"
  [(set (cc0) 
	(unspec:SI [(match_dup 1) (match_dup 2)] 1))
   (set (pc) (if_then_else (match_op_dup 3
			    [(cc0) (const_int 0)])
	     (label_ref (match_dup 4))
	     (pc)))]
"")

(define_peephole2
 [(set (match_operand:SI 0 "register_operand" "")
       (match_operand:SI 1 "nonimmediate_operand_msp430" ""))
  (set (match_dup 0) (and:SI (match_dup 0) 
                             (match_operand 2 "const_int_operand" "")))
  (set (pc) (if_then_else (match_operator:SI 3 "equality_operator"
			   [(match_dup 0) (const_int 0)])
	    (label_ref (match_operand 4 "" ""))
	    (pc)))]
"(halfnibble_integer(operands[2], VOIDmode) 
   || halfnibble_constant(operands[2], VOIDmode))
        && dead_or_set_in_peep(2,insn, operands[0])
	&& which_nibble(INTVAL(operands[2])) == 1"
  [(set (cc0) 
	(unspec:SI [(match_dup 1) (match_dup 2)] 2))
   (set (pc) (if_then_else (match_op_dup 3
			    [(cc0) (const_int 0)])
	     (label_ref (match_dup 4))
	     (pc)))]
"")


(define_peephole2
 [(set (match_operand:SI 0 "register_operand" "")
       (match_operand:SI 1 "nonimmediate_operand_msp430" ""))
  (set (match_dup 0) (and:SI (match_dup 0) 
                             (match_operand 2 "const_int_operand" "")))
  (set (pc) 
       (if_then_else (match_operator:HI 3 "equality_operator"
		      [(match_operand:HI 4 "register_operand" "") (const_int 0)]) 
       (label_ref (match_operand 5 "" ""))
       (pc)))]
"(halfnibble_integer(operands[2], VOIDmode) 
   || halfnibble_constant(operands[2], VOIDmode))
        && dead_or_set_in_peep(2,insn, operands[0])
	&& which_nibble(INTVAL(operands[2])) == 1
	&& REGNO(operands[4]) == REGNO(operands[0])+1"
  [(set (cc0) 
	(unspec:SI [(match_dup 1) (match_dup 2)] 1))
   (set (pc) (if_then_else (match_op_dup 3
			    [(cc0) (const_int 0)])
	     (label_ref (match_dup 5))
	     (pc)))]
"")

(define_peephole2
 [(set (match_operand:SI 0 "register_operand" "")
       (match_operand:SI 1 "nonimmediate_operand_msp430" ""))
  (set (match_dup 0) (and:SI (match_dup 0) 
                             (match_operand 2 "const_int_operand" "")))
  (set (pc) 
       (if_then_else (match_operator:HI 3 "equality_operator"
		      [(match_operand:HI 4 "register_operand" "") (const_int 0)]) 
       (label_ref (match_operand 5 "" ""))
       (pc)))]
"(halfnibble_integer(operands[2], VOIDmode) 
   || halfnibble_constant(operands[2], VOIDmode))
        && dead_or_set_in_peep(2,insn, operands[0])
	&& which_nibble(INTVAL(operands[2])) == 0
	&& REGNO(operands[4]) == REGNO(operands[0])"
  [(set (cc0) 
	(unspec:SI [(match_dup 1) (match_dup 2)] 1))
   (set (pc) (if_then_else (match_op_dup 3
			    [(cc0) (const_int 0)])
	     (label_ref (match_dup 5))
	     (pc)))]
"")

;;
;;  The same for HI mode: while(smts&0xXXXX) ;
;;
(define_insn "*bittest"
  [(set (cc0)
        (unspec:HI [(match_operand:HI 0 "general_operand_msp430" "r,r,m,m")
                    (match_operand:HI 1 "general_operand_msp430" "rPR,mi,rPR,mi")] 6))]
""
"bit	%1,%0"
[(set_attr "length" "1,2,2,3")
   (set_attr "cc" "compare,compare,compare,compare")])


(define_peephole2
 [(set (match_operand:HI 0 "register_operand" "")
       (match_operand:HI 1 "nonimmediate_operand_msp430" ""))
  (set (match_dup 0) (and:HI (match_dup 0) 
                             (match_operand 2 "const_int_operand" "")))
  (set (pc) (if_then_else (match_operator:HI 3 "equality_operator"
			   [(match_dup 0) (const_int 0)])
	    (label_ref (match_operand 4 "" "")) 
	    (pc)))]
"dead_or_set_in_peep(2,insn, operands[0]) "
  [(set (cc0) 
	(unspec:HI [(match_dup 1) (match_dup 2)] 6))
   (set (pc) (if_then_else (match_op_dup 3
			    [(cc0) (const_int 0)])
	     (label_ref (match_dup 4))
	     (pc)))]
"")


;; The same for QI mode

(define_insn "*bittest_b"
  [(set (cc0)
        (unspec:QI [(match_operand:QI 0 "general_operand_msp430" "r,r,m,m")
                    (match_operand:QI 1 "general_operand_msp430" "rPR,mi,rPR,mi")] 7))]
""
"bit.b	%1,%0"
[(set_attr "length" "1,2,2,3")
   (set_attr "cc" "compare,compare,compare,compare")])


(define_peephole2
 [(set (match_operand:QI 0 "register_operand" "")
       (match_operand:QI 1 "nonimmediate_operand_msp430" ""))
  (set (match_dup 0) (and:QI (match_dup 0) 
                             (match_operand 2 "const_int_operand" "")))
  (set (pc) (if_then_else (match_operator:QI 3 "equality_operator"
			   [(match_dup 0) (const_int 0)])
	    (label_ref (match_operand 4 "" "")) 
	    (pc)))]
"dead_or_set_in_peep(2,insn, operands[0]) "
  [(set (cc0) 
	(unspec:QI [(match_dup 1) (match_dup 2)] 7))
   (set (pc) (if_then_else (match_op_dup 3
			    [(cc0) (const_int 0)])
	     (label_ref (match_dup 4))
	     (pc)))]
"")



;;===========================================================================

(define_peephole2
  [(set (match_operand:QI 0 "register_operand" "")
        (match_operand:QI 1 "general_operand_msp430" ""))
   (set (match_dup 0)
	(minus:QI (match_dup 0) 
		  (match_operand:QI 2 "general_operand_msp430" "")))
   (set (match_operand:QI 3 "register_operand" "")
   	(match_dup 0))]
"dead_or_set_in_peep(2,insn, operands[0]) && 0"
  [(set (match_dup 3) (match_dup 1))
   (set (match_dup 3) (minus:QI (match_dup 3) (match_dup 2)))]
"")


(define_peephole2
  [(set (match_operand:QI 0 "register_operand" "")
        (match_operand:QI 1 "general_operand_msp430" ""))
   (set (match_dup 0)
	(plus:QI (match_dup 0) 
		 (match_operand:QI 2 "general_operand_msp430" "")))
   (set (match_operand:QI 3 "register_operand" "")
   	(match_dup 0))]
"dead_or_set_in_peep(2,insn, operands[0]) && 0"
  [(set (match_dup 3) (match_dup 1))
   (set (match_dup 3) (plus:QI (match_dup 3) (match_dup 2)))]
"")


(define_peephole2
  [(set (match_operand:HI 0 "register_operand" "")
        (match_operand:HI 1 "general_operand_msp430" ""))
   (set (match_dup 0)
	(minus:HI (match_dup 0) 
		  (match_operand:HI 2 "general_operand_msp430" "")))
   (set (match_operand:HI 3 "register_operand" "")
   	(match_dup 0))]
"dead_or_set_in_peep(2,insn, operands[0]) && 0"
  [(set (match_dup 3) (match_dup 1))
   (set (match_dup 3) (minus:HI (match_dup 3) (match_dup 2)))]
"")


(define_peephole2
  [(set (match_operand:HI 0 "register_operand" "")
        (match_operand:HI 1 "general_operand_msp430" ""))
   (set (match_dup 0)
	(plus:HI (match_dup 0) 
		 (match_operand:HI 2 "general_operand_msp430" "")))
   (set (match_operand:HI 3 "register_operand" "")
   	(match_dup 0))]
"dead_or_set_in_peep(2,insn, operands[0]) && 0"
  [(set (match_dup 3) (match_dup 1))
   (set (match_dup 3) (plus:HI (match_dup 3) (match_dup 2)))]
"")


(define_peephole2
  [(set (match_operand:SI 0 "register_operand" "")
        (match_operand:SI 1 "general_operand_msp430" ""))
   (set (match_dup 0)
	(minus:SI (match_dup 0) 
		  (match_operand:SI 2 "general_operand_msp430" "")))
   (set (match_operand:SI 3 "register_operand" "")
   	(match_dup 0))]
"dead_or_set_in_peep(2,insn, operands[0]) && 0"
  [(set (match_dup 3) (match_dup 1))
   (set (match_dup 3) (minus:SI (match_dup 3) (match_dup 2)))]
"")


(define_peephole2
  [(set (match_operand:SI 0 "register_operand" "")
        (match_operand:SI 1 "general_operand_msp430" ""))
   (set (match_dup 0)
	(plus:SI (match_dup 0) 
		 (match_operand:SI 2 "general_operand_msp430" "")))
   (set (match_operand:SI 3 "register_operand" "")
   	(match_dup 0))]
"dead_or_set_in_peep(2,insn, operands[0]) && 0"
  [(set (match_dup 3) (match_dup 1))
   (set (match_dup 3) (plus:SI (match_dup 3) (match_dup 2)))]
"")


;; =============================================================
;; 
;;  adjust frame pointer index
;;
(define_peephole2
  [(set (match_operand:HI 0 "register_operand" "") 
	(sign_extend:HI (match_operand:QI 1 "general_operand_msp430" "")))
   (set (match_dup 0) 
	(plus:HI (match_dup 0) (match_operand:HI 2 "general_operand_msp430" "")))
   (set (match_operand:HI 3 "register_operand" "")
	(match_dup 0))]
"dead_or_set_in_peep(2,insn, operands[0])"
  [(set (match_dup 3) (sign_extend:HI (match_dup 1)))
   (set (match_dup 3) (plus:HI (match_dup 3) (match_dup 2)))]
"")

(define_peephole2
  [(set (match_operand:SI 0 "register_operand" "") 
	(sign_extend:SI (match_operand:HI 1 "general_operand_msp430" "")))
   (set (match_dup 0) 
	(plus:SI (match_dup 0) (match_operand:SI 2 "general_operand_msp430" "")))
   (set (match_operand:SI 3 "register_operand" "")
	(match_dup 0))]
"dead_or_set_in_peep(2,insn, operands[0])"
  [(set (match_dup 3) (sign_extend:SI (match_dup 1)))
   (set (match_dup 3) (plus:SI (match_dup 3) (match_dup 2)))]
"")

(define_peephole2
  [(set (match_operand:HI 0 "register_operand" "") 
	(sign_extend:HI (match_operand:QI 1 "general_operand_msp430" "")))
   (set (match_dup 0) 
	(minus:HI (match_dup 0) (match_operand:HI 2 "general_operand_msp430" "")))
   (set (match_operand:HI 3 "register_operand" "")
	(match_dup 0))]
"dead_or_set_in_peep(2,insn, operands[0])"
  [(set (match_dup 3) (sign_extend:HI (match_dup 1)))
   (set (match_dup 3) (minus:HI (match_dup 3) (match_dup 2)))]
"")

(define_peephole2
  [(set (match_operand:SI 0 "register_operand" "") 
	(sign_extend:SI (match_operand:HI 1 "general_operand_msp430" "")))
   (set (match_dup 0) 
	(minus:SI (match_dup 0) (match_operand:SI 2 "general_operand_msp430" "")))
   (set (match_operand:SI 3 "register_operand" "")
	(match_dup 0))]
"dead_or_set_in_peep(2,insn, operands[0])"
  [(set (match_dup 3) (sign_extend:SI (match_dup 1)))
   (set (match_dup 3) (minus:SI (match_dup 3) (match_dup 2)))]
"")

;; =============================================================
;; mov & 'and'

(define_peephole2
  [(set (match_operand:HI 0 "register_operand" "") 
	(match_operand:HI 1 "nonimmediate_operand_msp430" ""))
   (set (match_dup 0) 
	(and:HI (match_dup 0) 
		(match_operand:HI 2 "general_operand_msp430" ""))) 
   (set (match_dup 1) (match_dup 0))]
"dead_or_set_in_peep(2,insn, operands[0])"
  [(set (match_dup 1) 
	(and:HI (match_dup 1) (match_dup 2)))]
"")


(define_peephole2
  [(set (match_operand:SI 0 "register_operand" "") 
	(match_operand:SI 1 "nonimmediate_operand_msp430" ""))
   (set (match_dup 0) 
	(and:SI (match_dup 0) 
		(match_operand:SI 2 "general_operand_msp430" ""))) 
   (set (match_dup 1) (match_dup 0))]
"dead_or_set_in_peep(2,insn, operands[0])"
  [(set (match_dup 1) 
	(and:SI (match_dup 1) (match_dup 2)))]
"")



;; =============================================================
;; SWAP BYTES (should be a pattern for:
;;   r = (a<<8)|(a>>8);

(define_insn "swpb"
  [(set (match_operand:HI 0 "nonimmediate_operand_msp430" "=rR,m") 
	(unspec:HI [(match_operand:HI 1 "nonimmediate_operand_msp430" "0,0")] 4))]
""
"swpb	%0"
[(set_attr "length" "1,2")
   (set_attr "cc" "clobber,clobber")])

;;
;; after mult and stuff
;;
(define_peephole2
  [(set (match_operand:SI 0 "register_operand" "") 
	(reg:SI 14))
   (set (match_operand:SI 1 "nonimmediate_operand_msp430" "") 
	(match_dup 0))]
"dead_or_set_in_peep(1,insn,operands[0])"
  [(set (match_dup 1) (reg:SI 14))]
"")

(define_peephole2
  [(set (match_operand:HI 0 "register_operand" "") 
	(reg:HI 14))
   (set (match_operand:HI 1 "nonimmediate_operand_msp430" "") 
	(match_dup 0))]
"dead_or_set_in_peep(1,insn,operands[0])"
  [(set (match_dup 1) (reg:HI 14))]
"")

(define_peephole2
  [(set (match_operand:SI 0 "register_operand" "") 
	(reg:SI 12))
   (set (match_operand:SI 1 "nonimmediate_operand_msp430" "") 
	(match_dup 0))]
"dead_or_set_in_peep(1,insn,operands[0])"
  [(set (match_dup 1) (reg:SI 12))]
"")

(define_peephole2
  [(set (match_operand:HI 0 "register_operand" "") 
	(reg:HI 12))
   (set (match_operand:HI 1 "nonimmediate_operand_msp430" "") 
	(match_dup 0))]
"dead_or_set_in_peep(1,insn,operands[0])"
  [(set (match_dup 1) (reg:HI 12))]
"")

(define_peephole2
  [(set (match_operand:SI 0 "register_operand" "") 
	(sign_extend:SI (match_operand:QI 1 "general_operand_msp430" "")))
   (set (match_operand:SI 2 "nonimmediate_operand_msp430" "") 
	(match_dup 0))]
"dead_or_set_in_peep(1,insn,operands[0])"
  [(set (match_dup 2) (sign_extend:SI (match_dup 1)))]
"")

(define_peephole2
  [(set (match_operand:SI 0 "register_operand" "") 
	(sign_extend:SI (match_operand:HI 1 "general_operand_msp430" "")))
   (set (match_operand:SI 2 "nonimmediate_operand_msp430" "") 
	(match_dup 0))]
"dead_or_set_in_peep(1,insn,operands[0])"
  [(set (match_dup 2) (sign_extend:SI (match_dup 1)))]
"")


;; =============================================================

(define_peephole 
  [(set (match_operand:HI 0 "register_operand" "") 
	(sign_extend:HI (match_operand:QI 1 "register_operand" ""))) 
   (set (match_operand:HI 2 "register_operand" "") 
	(plus:HI (match_dup 2) (match_dup 0)))]
"dead_or_set_in_peep(1,insn,operands[0])" 
"sxt	%1
	add	%1, %2"
[(set_attr "length" "3")
   (set_attr "cc" "clobber")])


(define_peephole 
  [(set (match_operand:HI 0 "register_operand" "") 
	(sign_extend:HI (match_operand:QI 1 "nonimmediate_operand_msp430" ""))) 
   (set (match_operand:HI 2 "register_operand" "") (match_dup 0))]
"dead_or_set_in_peep(1,insn,operands[0])" 
"mov.b	%1, %2
	sxt	%2"
[(set_attr "length" "2")
   (set_attr "cc" "clobber")])

;; ============================================================= 
;; a = (uint16_t)( (uint8_t)SFR ) << 8;
;; (inderect_jump + 50)
(define_peephole 
  [(set (match_operand:QI 0 "register_operand" "") 
	(match_operand:QI 1 "memory_operand_msp430" "m"))
   (set (match_operand:HI 2 "register_operand" "")
	(ashift:HI (match_dup 2) (const_int 8)))]
  "REGNO(operands[0]) == REGNO(operands[2])"
"mov.b	%1, %0
	swpb	%2"
[(set_attr "length" "3")
   (set_attr "cc" "clobber")])

(define_peephole2
  [(set (match_operand:HI 0 "register_operand" "") 
	(ior:HI (match_dup 0) 
		(match_operand:HI 1 "register_operand" "")))
   (set (match_operand:SI 2 "register_operand" "") 
	(match_operand:SI 3 "register_operand" ""))]
"(REGNO(operands[0]) == REGNO(operands[2])
  && REGNO(operands[3])+1 == REGNO(operands[0]))"
  [(set (match_dup 1) 
	(ior:HI (match_dup 1) 
		(match_dup 0)))
   (set (subreg:HI (match_dup 2) 0)
	(subreg:HI (match_dup 3) 0))]
"")



;; =============================================================
;; combine ior and mov.
;;
(define_peephole2
[(set (match_operand:QI 0 "register_operand" "") 
      (ior:QI (match_dup 0) 
      	      (match_operand:QI 1 "nonimmediate_operand_msp430" "")))
 (set (match_dup 1) (match_dup 0))]
"dead_or_set_in_peep(1,insn,operands[0])"
[(set (match_dup 1) (ior:QI (match_dup 1) (match_dup 0)))]
"")

(define_peephole2
[(set (match_operand:HI 0 "register_operand" "") 
      (ior:HI (match_dup 0) 
      	      (match_operand:HI 1 "nonimmediate_operand_msp430" "")))
 (set (match_dup 1) (match_dup 0))]
"dead_or_set_in_peep(1,insn,operands[0])"
[(set (match_dup 1) (ior:HI (match_dup 1) (match_dup 0)))]
"")

(define_peephole2
[(set (match_operand:SI 0 "register_operand" "") 
      (ior:SI (match_dup 0) 
      	      (match_operand:SI 1 "nonimmediate_operand_msp430" "")))
 (set (match_dup 1) (match_dup 0))]
"dead_or_set_in_peep(1,insn,operands[0])"
[(set (match_dup 1) (ior:SI (match_dup 1) (match_dup 0)))]
"")

(define_peephole2
[(set (match_operand:DI 0 "register_operand" "") 
      (ior:DI (match_dup 0) 
      	      (match_operand:DI 1 "nonimmediate_operand_msp430" "")))
 (set (match_dup 1) (match_dup 0))]
"dead_or_set_in_peep(1,insn,operands[0])"
[(set (match_dup 1) (ior:DI (match_dup 1) (match_dup 0)))]
"")


;; ======================================================================

;; Enable Interrupts
(define_insn "enable_interrupt"
  [(unspec [(const_int 0)] UNSPEC_EINT)]
  ""
  "eint"
  [(set_attr "length" "1")
  (set_attr "cc" "none")
  ])

;; Disable Interrupts
(define_insn "disable_interrupt"
  [(unspec [(const_int 0)] UNSPEC_DINT)]
  ""
  "dint"
  [(set_attr "length" "1")
  (set_attr "cc" "none")
  ])


;; return insn
(define_insn "return"
  [(return)]
  "reload_completed && msp430_empty_epilogue()"
  "* return msp430_emit_return(insn, operands, NULL);"
  [(set_attr "length" "1")
   (set_attr "cc" "clobber")]
)

(define_insn "return_from_epilogue"
  [(return)]
  "(reload_completed 
    && cfun->machine 
    && !(cfun->machine->is_interrupt || cfun->machine->is_critical)
    && !cfun->machine->is_naked)"
  "ret"
  [(set_attr "cc" "none")
   (set_attr "length" "1")])

(define_insn "return_from_interrupt_epilogue"
  [(return)]
  "(reload_completed 
    && cfun->machine 
    && (cfun->machine->is_interrupt || cfun->machine->is_critical)
    && !cfun->machine->is_naked)"
  "reti"
  [(set_attr "cc" "none")
   (set_attr "length" "1")])

(define_insn "return_from_naked_epilogue"
  [(return)]
  "(reload_completed 
    && cfun->machine 
    && cfun->machine->is_naked)"
  ""
  [(set_attr "cc" "none")
   (set_attr "length" "0")])

(define_expand "prologue"
  [(const_int 0)]
  ""
  "
  {
    expand_prologue (); 
    DONE;
  }")

(define_expand "epilogue"
  [(const_int 0)]
  ""
  "
  {
    expand_epilogue (); 
    DONE;
  }")
